# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""License Manager."""

from typing import List, Optional, Union

from ansys.mechanical.core import LOG


class LicenseManager:
    """Class to manage licenses in Ansys Mechanical.

    This class provides methods to enable, disable, and check the status of licenses.
    It also allows for moving licenses to specific indices in the license preference list.
    It is initialized with an instance of the Ansys Mechanical application.
    """

    def __init__(self, app):
        """Initialize the message manager."""
        self._app = app
        self._license_preference = None
        import Ansys

        self._license_status = Ansys.Mechanical.DataModel.Enums.LicenseStatus

    def _get_license_manager(self):
        if self._license_preference is None:
            self._license_preference = self._app.ExtAPI.Application.LicensePreference
        return self._license_preference

    def get_all_licenses(self) -> list[str]:
        """Return list of all licenses."""
        return self._get_license_manager().GetAllLicenses()

    def get_license_status(
        self, license_name: str
    ) -> "Ansys.Mechanical.DataModel.Enums.LicenseStatus":  # noqa: F821
        """Return status of the specific license.

        Parameters
        ----------
        license_name : str
            Name of the license to check.

        Returns
        -------
        "Ansys.Mechanical.DataModel.Enums.LicenseStatus"
            The status of the license.
        """
        LOG.info(
            f"{license_name} status: {self._get_license_manager().GetLicenseStatus(license_name)}"
        )
        return self._get_license_manager().GetLicenseStatus(license_name)

    def set_license_status(self, license_name: str, status: bool) -> None:
        """Set the status of a license and save the preference.

        Parameters
        ----------
        license_name : str
            Name of the license to set the status for.
        status : bool
            True to enable the license, False to disable it.
        """
        if status:
            self._get_license_manager().SetLicenseStatus(license_name, self._license_status.Enabled)
            LOG.info(f"{license_name} is enabled.")
        else:
            self._get_license_manager().SetLicenseStatus(
                license_name, self._license_status.Disabled
            )
            LOG.info(f"{license_name} is disabled.")
        self._get_license_manager().Save()

    def show(self) -> None:
        """Print all active licenses."""
        for lic in self.get_all_licenses():
            print(f"{lic} - {self._get_license_manager().GetLicenseStatus(lic)}")

    def disable_session_license(self) -> None:
        """Disable active license for current session."""
        self._get_license_manager().DeActivateLicense()

    def enable_session_license(self, license: Optional[Union[str, List[str]]] = None) -> None:
        """Enable license(s) for the current session.

        Parameters
        ----------
        license : Optional[Union[str, List[str]]], optional
            If None, activates the first enabled license in the priority order.
            If a string, activates that specific license.
            If a list of strings, activates all specified licenses in the order provided.
        """
        from System import String
        from System.Collections.Generic import List as DotNetList

        if license is None:
            self._get_license_manager().ActivateLicense()
        elif isinstance(license, str):
            self._get_license_manager().ActivateLicense(String(license))
        elif isinstance(license, list):
            licenses = DotNetList[String]()
            for lic in license:
                licenses.Add(String(lic))
            self._get_license_manager().ActivateLicense(licenses)
        else:
            raise TypeError("License must be None, a string, or a list of strings.")
        LOG.info(f"License(s) {license} enabled for the current session.")

    def move_to_index(self, license_name: str, location: int) -> None:
        """Move a license preference.

        Move license to zero-based index location in the license preference list.
        This is useful for setting the preferred license location.

        Parameters
        ----------
        license_name : str
            License name.
        location : int
            Location to move the license to.

        Examples
        --------
        Move Ansys Mechanical Premium to the first location.

        >>> license_manager = LicenseManager(app)
        >>> license_manager.move_to_index("Ansys Mechanical Premium", 0)
        """
        LOG.info(f"Moving license preference for {license_name} to location {location}")
        self._get_license_manager().MoveLicenseToLocation(license_name, location)
        self._get_license_manager().Save()

    def reset_preference(self) -> None:
        """Reset the license preference.

        This method will reset the license location order and the status of all licenses
        to the default state.
        """
        LOG.info("Resetting license preferences to default state.")
        self._get_license_manager().Reset()



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Clean up temporary mechdb files after GUI is closed."""

from pathlib import Path
import shutil
import sys
import time

import psutil


def cleanup_gui(pid, temp_mechdb) -> None:
    """Remove the temporary mechdb file after it is closed.

    Parameters
    ----------
    pid: int
        The process ID of the open temporary mechdb file.
    temp_mechdb: Path
        The path of the temporary mechdb file.
    """
    # While the pid exists, sleep
    while psutil.pid_exists(pid):
        time.sleep(1)

    # Delete the temporary mechdb file once the GUI is closed
    temp_mechdb.unlink()

    # Delete the temporary mechdb Mech_Files folder
    temp_folder_path = temp_mechdb.parent / f"{temp_mechdb.name.split('.')[0]}_Mech_Files"
    shutil.rmtree(temp_folder_path)


if __name__ == "__main__":  # pragma: no cover
    """Get the process ID and temporary file path to monitor and delete files after use."""
    # Convert the process id (pid) argument into an integer
    pid = int(sys.argv[1])
    # Convert the temporary mechdb path into a Path
    temp_mechdb_path = Path(sys.argv[2])
    # Remove the temporary mechdb file when the GUI is closed
    cleanup_gui(pid, temp_mechdb_path)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Run Mechanical UI from Python."""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.mechanical.core.embedding import App
from pathlib import Path

# Subprocess is needed to launch the GUI and clean up the process on close.
# Excluding bandit check.
from subprocess import Popen  # nosec: B404
import sys
import tempfile
import typing


class UILauncher:
    """Launch the GUI using a temporary mechdb file."""

    def __init__(self, dry_run: bool = False):
        """Initialize UILauncher class."""
        self._dry_run = dry_run

    def save_original(self, app: App) -> None:
        """Save the active mechdb file.

        Parameters
        ----------
        app: ansys.mechanical.core.embedding.app.App
            A Mechanical embedding application.
        """
        app.save()

    def save_temp_copy(self, app: App) -> typing.Union[Path, Path]:
        """Save a new mechdb file with a temporary name.

        Parameters
        ----------
        app: ansys.mechanical.core.embedding.app.App
            A Mechanical embedding application.
        """
        # Identify the mechdb of the saved session from save_original()
        project_directory = Path(app.project_directory)
        project_directory_parent = project_directory.parent
        mechdb_file = (
            project_directory_parent / f"{project_directory.parts[-1].split('_')[0]}.mechdb"
        )

        # Get name of NamedTemporaryFile
        temp_file_name = tempfile.NamedTemporaryFile(
            dir=project_directory_parent, suffix=".mechdb", delete=True
        ).name

        # Save app with name of temporary file
        app.save_as(temp_file_name)

        return mechdb_file, temp_file_name

    def open_original(self, app: App, mechdb_file: str) -> None:
        """Open the original mechdb file from save_original().

        Parameters
        ----------
        app: ansys.mechanical.core.embedding.app.App
            A Mechanical embedding application.
        mechdb_file: str
            The full path to the active mechdb file.
        """
        app.open(mechdb_file)

    def graphically_launch_temp(self, app: App, temp_file: Path) -> typing.Union[Popen, str]:
        """Launch the GUI for the mechdb file with a temporary name from save_temp_copy().

        Parameters
        ----------
        app: ansys.mechanical.core.embedding.app.App
            A Mechanical embedding application.
        temp_file: pathlib.Path
            The full path to the temporary mechdb file.

        Returns
        -------
        subprocess.Popen
            The subprocess that launches the GUI for the temporary mechdb file.
        """
        # The ansys-mechanical command to launch the GUI in a subprocess
        args = [
            "ansys-mechanical",
            "--project-file",
            temp_file,
            "--graphical",
            "--revision",
            str(app.version),
        ]
        if not self._dry_run:
            # The subprocess that uses ansys-mechanical to launch the GUI of the temporary
            # mechdb file
            process = Popen(args)  # nosec: B603 # pragma: no cover
            return process
        else:
            # Return a string containing the args
            return " ".join(args)

    def _cleanup_gui(self, process: Popen, temp_mechdb_path: Path) -> None:
        """Remove the temporary mechdb file and folder when the GUI is closed.

        Parameters
        ----------
        process: subprocess.Popen
            The subprocess that launched the GUI of the temporary mechdb file.
        temp_mechdb_path: pathlib.Path
            The full path to the temporary mechdb file.
        """
        # Get the path to the cleanup script
        cleanup_script = Path(__file__).parent / "cleanup_gui.py"  # pragma: no cover

        if not self._dry_run:
            # Open a subprocess to remove the temporary mechdb file and folder when the process ends
            Popen(
                [sys.executable, cleanup_script, str(process.pid), temp_mechdb_path]
            )  # pragma: no cover # nosec: B603


def _is_saved(app: App) -> bool:
    """Check if the mechdb file has been saved and raise an exception if not.

    Parameters
    ----------
    app: ansys.mechanical.core.embedding.app.App
        A Mechanical embedding application.

    Returns
    -------
    bool
        ``True`` when the embedded app has been saved.
        ``False`` when the embedded app has not been saved.
    """
    try:
        app.save()
    except Exception as e:
        raise Exception("The App must have already been saved before using launch_ui!") from e
    return True


def _launch_ui(app: App, delete_tmp_on_close: bool, launcher: UILauncher) -> None:
    """Launch the Mechanical UI if the mechdb file has been saved.

    Parameters
    ----------
    app: ansys.mechanical.core.embedding.app.App
        A Mechanical embedding application.
    delete_tmp_on_close: bool
        Whether to delete the temporary mechdb file when the GUI is closed.
        By default, this is ``True``.
    launcher: UILauncher
        Launch the GUI using a temporary mechdb file.
    """
    if _is_saved(app):
        # Save the active mechdb file.
        launcher.save_original(app)
        # Save a new mechdb file with a temporary name.
        mechdb_file, temp_file = launcher.save_temp_copy(app)
        # Open the original mechdb file from save_original().
        launcher.open_original(app, str(mechdb_file))
        # Launch the GUI for the mechdb file with a temporary name from save_temp_copy().
        process = launcher.graphically_launch_temp(app, temp_file)

        # If it's a dry run and graphically_launch_temp returned a string, print the string
        if isinstance(process, str):
            print(process)

        # If the user wants the temporary file to be deleted and graphically_launch_temp
        # returned a process. By default, this is True
        if delete_tmp_on_close and not isinstance(process, str):
            # Remove the temporary mechdb file and folder when the GUI is closed.
            launcher._cleanup_gui(process, temp_file)  # pragma: no cover
        else:
            # Let the user know that the mechdb started above will not automatically get cleaned up
            print(
                f"""Opened a new mechanical session based on {mechdb_file} named {temp_file}.
PyMechanical will not delete it after use."""
            )


def launch_ui(
    app: App,
    delete_tmp_on_close: bool = True,
    dry_run: bool = False,
) -> None:
    """Launch the Mechanical UI.

    Precondition: Mechanical has to have already been saved
    Side effect: If Mechanical has ever been saved, it overwrites that save.

    Parameters
    ----------
    app: ansys.mechanical.core.embedding.app.App
        A Mechanical embedding application.
    delete_tmp_on_close: bool
        Whether to delete the temporary mechdb file when the GUI is closed.
        By default, this is ``True``.
    dry_run: bool
        Whether or not to launch the GUI. By default, this is ``False``.
    """
    _launch_ui(app, delete_tmp_on_close, UILauncher(dry_run))



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Temporary Appdata for Ansys Mechanical."""

from pathlib import Path
import shutil
import sys
import warnings


class UniqueUserProfile:
    """Create Unique User Profile (for AppData)."""

    def __init__(self, profile_name: str, copy_profile: bool = True, dry_run: bool = False):
        """Initialize UniqueUserProfile class."""
        self._default_profile = Path("~").expanduser()
        self._location = self._default_profile / "PyMechanical-AppData" / profile_name
        self._dry_run = dry_run
        self.copy_profile = copy_profile
        self.initialize()

    def initialize(self) -> None:
        """
        Initialize the new profile location.

        Args:
            copy_profile (bool): If False, the copy_profile method will be skipped.
        """
        if self._dry_run:
            return
        if self.exists():
            self.cleanup()
        self.mkdirs()
        if self.copy_profile:
            self.copy_profiles()

    def cleanup(self) -> None:
        """Cleanup unique user profile."""
        if self._dry_run:
            return

        # Remove the appdata directory if it exists
        shutil.rmtree(self.location, ignore_errors=True)

        if self.location.is_dir():
            warnings.warn(
                f"The `private appdata` option was used, but {self.location} was not removed"
            )

    @property
    def location(self) -> str:
        """Return the profile name."""
        return self._location

    def update_environment(self, env) -> None:
        """Set environment variables for new user profile."""
        home = self.location
        if "win" in sys.platform:
            appdata_dir = home / "AppData"
            appdata_local_temp = str(appdata_dir / "Local" / "Temp")

            env["USERPROFILE"] = str(home)
            env["APPDATA"] = str(appdata_dir / "Roaming")
            env["LOCALAPPDATA"] = str(appdata_dir / "Local")
            env["TMP"] = appdata_local_temp
            env["TEMP"] = appdata_local_temp
        elif "lin" in sys.platform:
            env["HOME"] = str(home)

    def exists(self) -> bool:
        """Check if unique profile name already exists."""
        return self.location.exists()

    def mkdirs(self) -> None:
        """Create a unique user profile & set up the directory tree."""
        self.location.mkdir(parents=True, exist_ok=True)
        if "win" in sys.platform:
            locs = ["AppData/Roaming", "AppData/Local", "Documents"]
        elif "lin" in sys.platform:
            locs = [".config", "temp/reports"]

        for loc in locs:
            dir_name = self.location / loc
            dir_name.mkdir(parents=True, exist_ok=True)

    def copy_profiles(self) -> None:
        """Copy current user directories into a new user profile."""
        if "win" in sys.platform:
            locs = ["AppData/Roaming/Ansys", "AppData/Local/Ansys"]
        elif "lin" in sys.platform:
            locs = [".mw/Application Data/Ansys", ".config/Ansys"]
        for loc in locs:
            default_profile_loc = self._default_profile / loc
            temp_appdata_loc = self.location / loc
            if default_profile_loc.exists():
                shutil.copytree(default_profile_loc, temp_appdata_loc)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for scheduling background work in IPython shells.

Interactive Python, or IPython, offers an enhanced Python shell. This module
schedules all the work of the IPython shell on a background thread, allowing
the Main thread to be used exclusively for the shell frontend. As a result,
user-defined functions can be executed during idle time between blocks.
"""

import queue
import threading
import time

from ansys.mechanical.core import LOG

try:
    from IPython import get_ipython
    from IPython.core.interactiveshell import InteractiveShell

    FROM_IPYTHON = get_ipython() is not None
    HAS_IPYTHON = True
except ImportError:
    FROM_IPYTHON = False
    HAS_IPYTHON = False

CODE_QUEUE = queue.Queue()
RESULT_QUEUE = queue.Queue()
SHUTDOWN_EVENT = threading.Event()
EXEC_THREAD = None
ORIGINAL_RUN_CELL = None
EXECUTION_THREAD_ID: int = None


def _idle_sleep():
    time.sleep(0.01)


DEFAULT_IDLE_HOOK = _idle_sleep


class ShellHooks:
    """IPython shell lifetime hooks."""

    def __init__(self):
        self._idle_hook: callable = DEFAULT_IDLE_HOOK
        self._start_hook: callable = None
        self._end_hook: callable = None

    @property
    def idle_hook(self) -> callable:
        """Function to call between IPython block executions."""
        return self._idle_hook

    @idle_hook.setter
    def idle_hook(self, value: callable) -> None:
        self._idle_hook = value

    @property
    def start_hook(self) -> callable:
        """Function to call at the start of the block thread."""
        return self._start_hook

    @start_hook.setter
    def start_hook(self, value: callable) -> None:
        self._start_hook = value

    @property
    def end_hook(self) -> callable:
        """Function to call when the shell is exited."""
        return self._end_hook

    @end_hook.setter
    def end_hook(self, value: callable) -> None:
        self._end_hook = value

    def start(self):
        """Signal handler when the shell starts."""
        if self.start_hook is not None:
            self._start_hook()
            self._start_hook = None

    def idle(self):
        """Signal handler when the shell is idle."""
        if self.idle_hook is not None:
            self._idle_hook()

    def end(self):
        """Signal handler when the shell ends."""
        if self.end_hook is not None:
            self._end_hook()
            self._end_hook = None


SHELL_HOOKS = ShellHooks()


def _exec_from_queue(shell) -> bool:
    """Worker function for ipython execution.

    Return whether to break out of loop
    """
    try:
        code = CODE_QUEUE.get_nowait()
    except queue.Empty:
        # call the idle hook
        try:
            SHELL_HOOKS.idle()
        except Exception as e:
            LOG.error(f"shell hook raised {e}. Uninstalling it.")
            SHELL_HOOKS.idle_hook = DEFAULT_IDLE_HOOK
        return False
    if code is None:
        return True
    LOG.info(f"execution thread {threading.get_ident()}")
    result = ORIGINAL_RUN_CELL(shell, code, store_history=True)
    RESULT_QUEUE.put(result)
    return False


def _execution_thread_main():
    global EXECUTION_THREAD_ID

    SHELL_HOOKS.start()
    shell = InteractiveShell.instance()
    EXECUTION_THREAD_ID = threading.get_ident()
    while not SHUTDOWN_EVENT.is_set():
        if _exec_from_queue(shell):
            break

    # one more execution in case the shutdown event was set and the exit code was not processed?
    if SHELL_HOOKS.end_hook is not None:
        _exec_from_queue(shell)


def _run_cell_in_thread(self, raw_cell, store_history=False, silent=False, shell_futures=True):
    CODE_QUEUE.put(raw_cell)
    while not SHUTDOWN_EVENT.is_set():
        try:
            return RESULT_QUEUE.get(timeout=0.1)
        except queue.Empty:
            continue
    raise RuntimeError("Execution thread shut down before result was returned.")


def cleanup():
    """Cleanup the ipython shell.

    Must be called before the application exits.
    May be called from an atexit handler.
    """
    LOG.info("Shutting down execution thread")
    CODE_QUEUE.put(None)  # Unblock the thread
    SHUTDOWN_EVENT.set()
    EXEC_THREAD.join(timeout=1)


def _can_post_ipython_blocks():
    if not HAS_IPYTHON:
        raise Exception("`post_ipython_blocks` requires ipython")


def in_ipython():
    """Return whether Python is running from IPython."""
    return FROM_IPYTHON


# Capture the original run_cell before patching
def post_ipython_blocks():
    """Initiate the IPython worker thread for block execution."""
    _can_post_ipython_blocks()
    global EXEC_THREAD
    global ORIGINAL_RUN_CELL
    LOG.info(f"original main thread {threading.get_ident()}")
    ORIGINAL_RUN_CELL = InteractiveShell.run_cell
    EXEC_THREAD = threading.Thread(target=_execution_thread_main, daemon=True)
    EXEC_THREAD.start()

    # Patch IPython to delegate to your thread
    InteractiveShell.run_cell = _run_cell_in_thread.__get__(
        InteractiveShell.instance(), InteractiveShell
    )
    LOG.info("IPython now runs all cells in your dedicated thread.")


def get_shell_hooks():
    """Get the shell hooks object."""
    return SHELL_HOOKS



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Main application class for embedded Mechanical."""

from __future__ import annotations

import atexit
import os
from pathlib import Path
import typing

from ansys.mechanical.core import LOG
from ansys.mechanical.core.embedding import initializer, runtime
from ansys.mechanical.core.embedding.addins import AddinConfiguration
from ansys.mechanical.core.embedding.appdata import UniqueUserProfile
from ansys.mechanical.core.embedding.imports import global_entry_points, global_variables
from ansys.mechanical.core.embedding.license_manager import LicenseManager
from ansys.mechanical.core.embedding.mechanical_warnings import (
    connect_warnings,
    disconnect_warnings,
)
from ansys.mechanical.core.embedding.poster import Poster
import ansys.mechanical.core.embedding.shell as shell
from ansys.mechanical.core.embedding.ui import launch_ui
from ansys.mechanical.core.feature_flags import get_command_line_arguments

if typing.TYPE_CHECKING:
    # Make sure to run ``ansys-mechanical-ideconfig`` to add the autocomplete settings to VS Code
    # Run ``ansys-mechanical-ideconfig --help`` for more information
    import Ansys  # pragma: no cover

try:
    import ansys.tools.visualization_interface  # noqa: F401

    HAS_ANSYS_GRAPHICS = True
    """Whether or not PyVista exists."""
except ImportError:
    HAS_ANSYS_GRAPHICS = False

shell.initialize_ipython_shell()


def _get_default_addin_configuration() -> AddinConfiguration:
    configuration = AddinConfiguration()
    return configuration


INSTANCES = []
"""List of instances."""


def _atexit_embedded_app(instances):  # pragma: nocover
    if len(instances) > 0:
        instance = instances[0]
        if instance._started_interactive_shell:
            shell.end_interactive_shell()
        else:
            instance._dispose()


def _cleanup_private_appdata(profile: UniqueUserProfile):
    profile.cleanup()


def _start_application(
    configuration: AddinConfiguration, version, db_file, _addtional_args
) -> "App":
    import clr

    clr.AddReference("Ansys.Mechanical.Embedding")
    import Ansys

    if configuration.no_act_addins:
        os.environ["ANSYS_MECHANICAL_STANDALONE_NO_ACT_EXTENSIONS"] = "1"

    addin_configuration_name = configuration.addin_configuration
    if version < 261:
        return Ansys.Mechanical.Embedding.Application(db_file, addin_configuration_name)
    return Ansys.Mechanical.Embedding.Application(
        db_file, addin_configuration_name, _addtional_args
    )


def _additional_args(readonly: bool, feature_flags: list, version: int) -> str:
    """Generate additional command line arguments for the application."""
    if version < 261:
        LOG.warning(
            "The readonly and feature_flags arguments are only supported "
            "with version 2026R1 and later."
        )
        return ""
    additional_args = ""
    if readonly:
        additional_args += " -readonly"
    if feature_flags:
        flag_args = get_command_line_arguments(feature_flags)
        additional_args += " " + " ".join(flag_args)
    return additional_args


def is_initialized():
    """Check if the app has been initialized."""
    return len(INSTANCES) != 0


class GetterWrapper(object):
    """Wrapper class around an attribute of an object."""

    def __init__(self, obj, getter):
        """Create a new instance of GetterWrapper."""
        # immortal class which provides wrapped object
        self.__dict__["_immortal_object"] = obj
        # function to get the wrapped object from the immortal class
        self.__dict__["_get_wrapped_object"] = getter

    def __getattr__(self, attr):
        """Wrap getters to the wrapped object."""
        if attr in self.__dict__:
            return getattr(self, attr)
        return getattr(self._get_wrapped_object(self._immortal_object), attr)

    def __setattr__(self, attr, value):
        """Wrap setters to the wrapped object."""
        if attr in self.__dict__:
            setattr(self, attr, value)
        setattr(self._get_wrapped_object(self._immortal_object), attr, value)


class App:
    """Mechanical embedding Application.

    Parameters
    ----------
    db_file : str, optional
        Path to a mechanical database file (.mechdat or .mechdb).
    version : int, optional
        Version number of the Mechanical application.
    private_appdata : bool, optional
        Setting for a temporary AppData directory. Default is False.
        Enables running parallel instances of Mechanical.
    globals : dict, optional
        Global variables to be updated. For example, globals().
        Replaces "app.update_globals(globals())".
    config : AddinConfiguration, optional
        Configuration for addins. By default "Mechanical" is used and ACT Addins are disabled.
    copy_profile : bool, optional
        Whether to copy the user profile when private_appdata is True. Default is True.
    enable_logging : bool, optional
        Whether to enable logging. Default is True.
    log_level : str, optional
        The logging level for the application. Default is "WARNING".
    pep8 : bool, optional
        Whether to enable PEP 8 style binding for the assembly. Default is False.
    readonly : bool, optional
        Whether to open the application in read-only mode. Default is False.
    feature_flags : list, optional
        List of feature flag names to enable. Default is [].
        Available flags include: 'ThermalShells', 'MultistageHarmonic', 'CPython'.

    Examples
    --------
    Create App with Mechanical project file and version:

    >>> from ansys.mechanical.core import App
    >>> app = App(db_file="path/to/file.mechdat", version=252)

    Disable copying the user profile when private appdata is enabled

    >>> app = App(private_appdata=True, copy_profile=False)

    Update the global variables with globals

    >>> app = App(globals=globals())

    Create App with "Mechanical" configuration and no ACT Addins

    >>> from ansys.mechanical.core.embedding import AddinConfiguration
    >>> from ansys.mechanical.core import App
    >>> config = AddinConfiguration("Mechanical")
    >>> config.no_act_addins = True
    >>> app = App(config=config)

    Set log level

    >>> app = App(log_level="INFO")

    ... INFO -  -  app - log_info - Starting Mechanical Application

    Create App in read-only mode

    >>> app = App(readonly=True)

    Create App with feature flags enabled

    >>> app = App(feature_flags=["CPython", "ThermalShells"])

    """

    def __init__(self, db_file=None, private_appdata=False, **kwargs):
        """Construct an instance of the mechanical Application."""
        global INSTANCES
        from ansys.mechanical.core import BUILDING_GALLERY

        self._enable_logging = kwargs.get("enable_logging", True)
        if self._enable_logging:
            self._log = LOG
            self._log_level = kwargs.get("log_level", "WARNING")
            self._log.setLevel(self._log_level)

        self.log_info("Starting Mechanical Application")

        # Get the globals dictionary from kwargs
        globals = kwargs.get("globals")

        # Get the interactive mode from kwargs
        self._interactive_mode = kwargs.get("interactive", False)

        # Set messages to None before BUILDING_GALLERY check
        self._messages = None

        version = kwargs.get("version")
        if version is not None:
            try:
                version = int(version)
            except ValueError:
                raise ValueError(
                    "The version must be an integer or of type that can be converted to an integer."
                )

        # If the building gallery flag is set, we need to share the instance
        # This can apply to running the `make -C doc html` command
        if BUILDING_GALLERY:
            if len(INSTANCES) != 0:
                # Get the first instance of the app
                instance: App = INSTANCES[0]
                # Point to the same underlying application object
                instance._share(self)
                # Update the globals if provided in kwargs
                if globals:
                    # The next line is covered by test_globals_kwarg_building_gallery
                    instance.update_globals(globals)  # pragma: nocover
                # Open the mechdb file if provided
                if db_file is not None:
                    self.open(db_file)
                return
        if len(INSTANCES) > 0:
            raise Exception("Cannot have more than one embedded mechanical instance!")

        self._version = initializer.initialize(version)

        if self._version < 261 and self.interactive:
            raise Exception("Interactive mode is only supported starting with version 261")

        configuration = kwargs.get("config", _get_default_addin_configuration())

        if private_appdata:
            copy_profile = kwargs.get("copy_profile", True)
            new_profile_name = f"PyMechanical-{os.getpid()}"
            profile = UniqueUserProfile(new_profile_name, copy_profile=copy_profile)
            profile.update_environment(os.environ)

        pep8_alias = kwargs.get("pep8", False)
        readonly = kwargs.get("readonly", False)
        feature_flags = kwargs.get("feature_flags", [])
        if readonly or feature_flags:
            additional_args = _additional_args(readonly, feature_flags, self._version)
        else:
            additional_args = ""

        self._prepare_interactive_mode()
        runtime.initialize(self._version, pep8_aliases=pep8_alias)

        self._app = _start_application(configuration, self._version, db_file, additional_args)

        self._disposed = False
        atexit.register(_atexit_embedded_app, INSTANCES)
        INSTANCES.append(self)

        self._handle_interactive_shell()

        connect_warnings(self)
        self._poster = None

        # Clean up the private appdata directory on exit if private_appdata is True
        if private_appdata:
            atexit.register(_cleanup_private_appdata, profile)

        self._updated_scopes: typing.List[typing.Dict[str, typing.Any]] = []
        self._subscribe()
        if globals:
            self.update_globals(globals)

        # Licensing API is available only for version 2025R2 and later
        self._license_manager = None
        if self.version >= 252:
            self._license_manager = LicenseManager(self)

    def __repr__(self):
        """Get the product info."""
        import clr

        clr.AddReference("Ansys.Mechanical.Application")
        import Ansys

        return Ansys.Mechanical.Application.ProductInfo.ProductInfoAsString

    def __enter__(self):  # pragma: no cover
        """Enter the scope."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        """Exit the scope."""
        self._dispose()

    def _dispose(self):
        if self._disposed:
            return
        self._unsubscribe()
        disconnect_warnings(self)
        self._app.Dispose()
        self._disposed = True

    def _prepare_interactive_mode(self):
        if not self.interactive:
            return
        if self._version < 261:
            raise Exception("Interactive mode is only supported starting with version 261")
        if os.name != "nt":
            if os.environ.get("ANSYS_MECHANICAL_EMBEDDING_LINUX_UI", False) is False:
                raise Exception("Interactive mode is only supported on windows")
        os.environ["ANSYS_MECHANICAL_EMBEDDING_START_WITH_UI"] = "1"

    def _handle_interactive_shell(self):
        self._started_interactive_shell = False
        if not self.interactive:
            return
        shell.start_interactive_shell(self)
        self._started_interactive_shell = True

    def wait_with_dialog(self):
        """Block python while an interactive pop-up is displayed.

        While that pop-up is displayed, Mechanical's main thread will
        not be blocked.
        """
        if not self.interactive:
            raise Exception("wait_with_dialog is only supported in interactive mode!")
        if self.version < 261:
            raise Exception("wait_with_dialog is only supported with Mechanical 261")
        # Wait with dialog open
        self._app.BlockingModalDialog("Wait with dialog", "PyMechanical")

    @property
    def interactive(self) -> bool:
        """Get whether the application is running in interactive mode."""
        return self._interactive_mode

    def open(self, db_file, remove_lock=False):
        """Open the db file.

        Parameters
        ----------
        db_file : str
            Path to a Mechanical database file (.mechdat or .mechdb).
        remove_lock : bool, optional
            Whether or not to remove the lock file if it exists before opening the project file.
        """
        self.log_info(f"Opening {db_file} ...")
        if remove_lock:
            lock_file = Path(self.DataModel.Project.ProjectDirectory) / ".mech_lock"
            # Remove the lock file if it exists before opening the project file
            if lock_file.exists():
                self.log_warning(
                    f"Removing the lock file, {lock_file}, before opening the project. "
                    "This may corrupt the project file."
                )
                lock_file.unlink()

        self.DataModel.Project.Open(db_file)

    def save(self, path=None):
        """Save the project."""
        if path is not None:
            self.DataModel.Project.Save(path)
        else:
            self.DataModel.Project.Save()

    def save_as(self, path: str, overwrite: bool = False, remove_lock: bool = False):
        """
        Save the project as a new file.

        If the `overwrite` flag is enabled, the current saved file is replaced with the new file.

        Parameters
        ----------
        path : str
            The path where the file needs to be saved.
        overwrite : bool, optional
            Whether the file should be overwritten if it already exists (default is False).
        remove_lock : bool, optional
            Whether to remove the lock file if it exists before saving the project file.

        Raises
        ------
        Exception
            If the file already exists at the specified path and `overwrite` is False.
        """
        if not Path(path).exists():
            self.DataModel.Project.SaveAs(path)
            return

        if not overwrite:
            raise Exception(
                f"File already exists in {path}, Use ``overwrite`` flag to "
                "replace the existing file."
            )

        if remove_lock:
            file_path = Path(path)
            associated_dir = file_path.parent / f"{file_path.stem}_Mech_Files"
            lock_file = associated_dir / ".mech_lock"
            # Remove the lock file if it exists before saving the project file
            if lock_file.exists():
                self.log_warning(f"Removing the lock file, {lock_file}... ")
                lock_file.unlink()
        try:
            self.DataModel.Project.SaveAs(path, overwrite)
        except Exception as e:
            error_msg = str(e)
            if "The project is locked by" in error_msg:
                self.log_error(
                    f"Failed to save project as {path}: {error_msg}\n"
                    "Hint: The project file is locked. "
                    "Try using the 'remove_lock=True' option when saving the project."
                )
            else:
                self.log_error(f"Failed to save project as {path}: {error_msg}")
            raise e

    def launch_gui(self, delete_tmp_on_close: bool = True, dry_run: bool = False):
        """Launch the GUI."""
        launch_ui(self, delete_tmp_on_close, dry_run)

    def new(self):
        """Clear to a new application."""
        self.DataModel.Project.New()

    def close(self):
        """Close the active project."""
        # Call New() to remove the lock file of the
        # current project on close.
        self.DataModel.Project.New()

    def exit(self):
        """Exit the application."""
        self._unsubscribe()
        self.ExtAPI.Application.Exit()

    def execute_script(self, script: str) -> typing.Any:
        """Execute the given script with the internal IronPython engine."""
        script_scope = "pymechanical-internal"
        if not hasattr(self, "script_engine"):
            import clr

            clr.AddReference("Ansys.Mechanical.Scripting")
            import Ansys

            # CreateEngine API without parameters creates an IronPython engine
            script_engine = Ansys.Mechanical.Scripting.EngineFactory.CreateEngine()
            empty_scope = False
            debug_mode = False
            script_engine.CreateScope(script_scope, empty_scope, debug_mode)
            self.script_engine = script_engine
        light_mode = True
        args = None
        rets = None
        script_result = self.script_engine.ExecuteCode(script, script_scope, light_mode, args, rets)
        error_msg = "Failed to execute the script"
        if script_result is None:
            raise Exception(error_msg)
        if script_result.Error is not None:
            error_msg += f": {script_result.Error.Message}"
            raise Exception(error_msg)
        return script_result.Value

    def execute_script_from_file(self, file_path=None):
        """Execute the given script from file with the internal IronPython engine."""
        file_path = Path(file_path)
        if not file_path.is_file():
            raise FileNotFoundError(f"The file {file_path} does not exist.")
        text_file = file_path.open("r", encoding="utf-8")
        data = text_file.read()
        text_file.close()
        return self.execute_script(data)

    def plotter(self, obj=None) -> None:
        """Return ``ansys.tools.visualization_interface.Plotter`` object."""
        if not HAS_ANSYS_GRAPHICS:
            LOG.warning(
                "Use ``pip install ansys-mechanical-core[graphics]`` to enable this option."
            )
            return

        if self.version < 242:
            LOG.warning("Plotting is only supported with version 2024R2 and later!")
            return

        # TODO : Check if anything loaded inside app or else show warning and return

        from ansys.mechanical.core.embedding.graphics.embedding_plotter import to_plotter

        return to_plotter(self, obj)

    def plot(self, obj=None) -> None:
        """Visualize the model in 3d.

        Requires installation using the graphics option. E.g.
        pip install ansys-mechanical-core[graphics]

        Examples
        --------
        >>> from ansys.mechanical.core import App
        >>> app = App()
        >>> app.open("path/to/file.mechdat")
        >>> app.plot()
        """
        if self.interactive:
            raise Exception("Plotting is not allowed in interactive mode")

        _plotter = self.plotter(obj)

        if _plotter is None:
            print("nothing to plot!")
            return

        return _plotter.show()

    @property
    def poster(self) -> Poster:
        """Returns an instance of Poster."""
        if self._poster is None:
            self._poster = Poster()
        return self._poster

    @property
    def DataModel(self) -> Ansys.Mechanical.DataModel.Interfaces.DataModelObject:  # noqa: N802
        """Return the DataModel."""
        return GetterWrapper(self._app, lambda app: app.DataModel)

    @property
    def ExtAPI(self) -> Ansys.ACT.Interfaces.Mechanical.IMechanicalExtAPI:  # noqa: N802
        """Return the ExtAPI object."""
        return GetterWrapper(self._app, lambda app: app.ExtAPI)

    @property
    def Tree(self) -> Ansys.ACT.Automation.Mechanical.Tree:  # noqa: N802
        """Return the Tree object."""
        return GetterWrapper(self._app, lambda app: app.DataModel.Tree)

    @property
    def Model(self) -> Ansys.ACT.Automation.Mechanical.Model:  # noqa: N802
        """Return the Model object."""
        return GetterWrapper(self._app, lambda app: app.DataModel.Project.Model)

    @property
    def Graphics(self) -> Ansys.ACT.Common.Graphics.MechanicalGraphicsWrapper:  # noqa: N802
        """Return the Graphics object."""
        return GetterWrapper(self._app, lambda app: app.ExtAPI.Graphics)

    @property
    def readonly(self):
        """Return whether the Mechanical object is read-only."""
        import Ansys

        return Ansys.ACT.Mechanical.MechanicalAPI.Instance.ReadOnlyMode

    @property
    def version(self):
        """Returns the version of the app."""
        return self._version

    @property
    def project_directory(self):
        """Returns the current project directory."""
        return self.DataModel.Project.ProjectDirectory

    @property
    def messages(self):
        """Lazy-load the MessageManager."""
        if self._messages is None:
            from ansys.mechanical.core.embedding.messages import MessageManager

            self._messages = MessageManager(self._app)
        return self._messages

    @property
    def license_manager(self):
        """Return license manager."""
        if self._license_manager is None:
            raise Exception("LicenseManager is only available for version 252 and later.")
        return self._license_manager

    def _share(self, other) -> None:
        """Shares the state of self with other.

        Other is another instance of App.
        This is used when the BUILDING_GALLERY flag is on.
        In that mode, multiple instance of App are used, but
        they all point to the same underlying application
        object. Because of that, special care needs to be
        taken to properly share the state. Other will be
        a "weak reference", which doesn't own anything.
        """
        # the other app is not expecting to have a project
        # already loaded
        self.new()

        # set up the type hint (typing.Self is python3.11+)
        other: App = other

        # copy `self` state to other.
        other._app = self._app
        other._version = self._version
        other._poster = self._poster
        other._updated_scopes = self._updated_scopes

        # all events will be handled by the original App instance
        other._subscribed = False

        # finally, set the other disposed flag to be true
        # so that the shutdown sequence isn't duplicated
        other._disposed = True

    def _subscribe(self):
        try:
            # This will throw an error when using pythonnet because
            # EventSource isn't defined on the IApplication interface
            self.ExtAPI.Application.EventSource.OnWorkbenchReady += self._on_workbench_ready
            self._subscribed = True
        except Exception:
            self._subscribed = False

    def _unsubscribe(self):
        if not self._subscribed:
            return
        self._subscribed = False
        self.ExtAPI.Application.EventSource.OnWorkbenchReady -= self._on_workbench_ready

    def _on_workbench_ready(self, sender, args) -> None:
        self._update_all_globals()

    def update_globals(
        self, globals_dict: typing.Dict[str, typing.Any], enums: bool = True
    ) -> None:
        """Update global variables.

        When scripting inside Mechanical, the Mechanical UI automatically
        sets global variables in Python. PyMechanical cannot do that automatically,
        but this method can be used.

        By default, all enums will be imported too. To avoid including enums, set
        the `enums` argument to False.

        Examples
        --------
        >>> from ansys.mechanical.core import App
        >>> app = App()
        >>> app.update_globals(globals())
        """
        self._updated_scopes.append(globals_dict)
        globals_dict.update(global_variables(self, enums))

    def _update_all_globals(self) -> None:
        for scope in self._updated_scopes:
            scope.update(global_entry_points(self))

    def _print_tree(self, node, max_lines, lines_count, indentation):
        """Recursively print till provided maximum lines limit.

        Each object in the tree is expected to have the following attributes:
         - Name: The name of the object.
         - Suppressed : Print as suppressed, if object is suppressed.
         - Children: Checks if object have children.
           Each child node is expected to have the all these attributes.

        Parameters
        ----------
        lines_count: int, optional
            The current count of lines printed. Default is 0.
        indentation: str, optional
            The indentation string used for printing the tree structure. Default is "".
        """
        if lines_count >= max_lines and max_lines != -1:
            print(f"... truncating after {max_lines} lines")
            return lines_count

        if not hasattr(node, "Name"):
            raise AttributeError("Object must have a 'Name' attribute")

        node_name = node.Name
        if hasattr(node, "Suppressed") and node.Suppressed is True:
            node_name += " (Suppressed)"
        if hasattr(node, "ObjectState"):
            if str(node.ObjectState) == "UnderDefined":
                node_name += " (?)"
            elif str(node.ObjectState) == "Solved" or str(node.ObjectState) == "FullyDefined":
                node_name += " (✓)"
            elif str(node.ObjectState) == "NotSolved" or str(node.ObjectState) == "Obsolete":
                node_name += " (⚡︎)"
            elif str(node.ObjectState) == "SolveFailed":
                node_name += " (✕)"
        print(f"{indentation}├── {node_name}")
        lines_count += 1

        if lines_count >= max_lines and max_lines != -1:
            print(f"... truncating after {max_lines} lines")
            return lines_count

        if hasattr(node, "Children") and node.Children is not None and node.Children.Count > 0:
            for child in node.Children:
                lines_count = self._print_tree(child, max_lines, lines_count, indentation + "|  ")
                if lines_count >= max_lines and max_lines != -1:
                    break

        return lines_count

    def print_tree(self, node=None, max_lines=80, lines_count=0, indentation=""):
        """
        Print the hierarchical tree representation of the Mechanical project structure.

        Parameters
        ----------
        node: DataModel object, optional
            The starting object of the tree.
        max_lines: int, optional
            The maximum number of lines to print. Default is 80. If set to -1, no limit is applied.

        Raises
        ------
        AttributeError
            If the node does not have the required attributes.

        Examples
        --------
        >>> from ansys.mechanical.core import App
        >>> app = App(globals=globals())
        >>> app.print_tree()
        ... ├── Project
        ... |  ├── Model
        ... |  |  ├── Geometry Imports (⚡︎)
        ... |  |  ├── Geometry (?)
        ... |  |  ├── Materials (✓)
        ... |  |  ├── Coordinate Systems (✓)
        ... |  |  |  ├── Global Coordinate System (✓)
        ... |  |  ├── Remote Points (✓)
        ... |  |  ├── Mesh (?)

        >>> app.print_tree(Model, 3)
        ... ├── Model
        ... |  ├── Geometry Imports (⚡︎)
        ... |  ├── Geometry (?)
        ... ... truncating after 3 lines

        >>> app.print_tree(max_lines=2)
        ... ├── Project
        ... |  ├── Model
        ... ... truncating after 2 lines
        """
        if node is None:
            node = self.DataModel.Project

        self._print_tree(node, max_lines, lines_count, indentation)

    def log_debug(self, message):
        """Log the debug message."""
        if not self._enable_logging:
            return
        self._log.debug(message)

    def log_info(self, message):
        """Log the info message."""
        if not self._enable_logging:
            return
        self._log.info(message)

    def log_warning(self, message):
        """Log the warning message."""
        if not self._enable_logging:
            return
        self._log.warning(message)

    def log_error(self, message):
        """Log the error message."""
        if not self._enable_logging:
            return
        self._log.error(message)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Cli for finding mechanical installation."""

import os
from pathlib import Path
import sys

import ansys.tools.path as atp
import click


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-r",
    "--version",
    default=None,
    type=int,
    help='Ansys version number, such as "242" or "241".\
         If a version number is not specified, it uses the default from \
            ansys-tools-path.',
)
@click.option(
    "-p",
    "--path",
    default=None,
    type=click.Path(exists=True, file_okay=False),
    help="Optional path to the Ansys installation directory. "
    "If provided, this path will be used instead of the default.",
)
def cli(version: int, path: str | None = None) -> tuple[int, str]:
    """
    Use the CLI tool to find the Mechanical version and location.

    Parameters
    ----------
    version: int
        Ansys version number.
    path: str, optional
        Optional path to the Ansys installation directory.
        eg: "usr/ansys_inc/v251/"

    Example
    -------
    Get the version and location of the installation directory.

    >>> find-mechanical -r 251
    >>> find-mechanical -p "usr/ansys_inc/v251/"
    """
    # Priority of finding the Mechanical version and path:
    # 1. If both path and version are provided, check if they match and use them
    # 2. If only path is provided, determine the version from the path
    # 3. If neither is provided, check environment variables for AWP_ROOT
    # 3. If multiple AWP_ROOT variables are found, use the highest version
    # 4. If no AWP_ROOT variables are found, use the default from ansys-tools-path
    # 5. If the default path is not found, check saved ansys path from config file
    if path and version:
        if version != atp.version_from_path("mechanical", path):
            raise click.BadParameter(
                f"The provided path {path} does not match the specified version {version}."
            )
        print(version, str(Path(path) / "aisol"))
        sys.exit(0)
    if path and not version:
        _version_from_given_path = atp.version_from_path("mechanical", path)
        print(_version_from_given_path, str(Path(path) / "aisol"))
        sys.exit(0)

    awp_roots = [value for key, value in os.environ.items() if key.startswith("AWP_ROOT")]

    if awp_roots:
        versions_found = []
        for awp_root in awp_roots:
            vfolder = Path(awp_root).name
            _version = vfolder.lstrip("v")
            versions_found.append(int(_version))
        if version in versions_found:
            print(version, str(Path(os.environ[f"AWP_ROOT{version}"]) / "aisol"))
            sys.exit(0)
        if versions_found:
            latest_version = max(versions_found)
            print(latest_version, str(Path(os.environ[f"AWP_ROOT{latest_version}"]) / "aisol"))
            sys.exit(0)

    # Use ansys-tools-path
    _exe = atp.get_mechanical_path(allow_input=False, version=version)
    _version = atp.version_from_path("mechanical", _exe)

    _path = str(Path(_exe).parent)
    print(_version, _path)

    sys.exit(0)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Class for running Mechanical on a background thread."""

import atexit
import threading
import time
import typing

import ansys.mechanical.core as mech
from ansys.mechanical.core.embedding import initializer
from ansys.mechanical.core.embedding.poster import Poster
import ansys.mechanical.core.embedding.utils as utils


def _exit(background_app: "BackgroundApp"):
    """Stop the thread serving the Background App."""
    background_app.stop()


class BackgroundApp:
    """Background App."""

    __app: mech.App = None
    __app_thread: threading.Thread = None
    __stopped: bool = False
    __stop_signaled: bool = False
    __poster: Poster = None

    def __init__(self, **kwargs):
        """Construct an instance of BackgroundApp."""
        if BackgroundApp.__app_thread is None:
            initializer.initialize(kwargs.get("version"))
            BackgroundApp.__app_thread = threading.Thread(
                target=self._start_app, kwargs=kwargs, daemon=True
            )
            BackgroundApp.__app_thread.start()

            while BackgroundApp.__poster is None:
                time.sleep(0.05)
                continue
        else:
            if BackgroundApp.__stopped:
                raise RuntimeError("Cannot initialize a BackgroundApp once it has been stopped!")

            def new():
                BackgroundApp.__app.new()

            self.post(new)

    @property
    def app(self) -> mech.App:
        """Get the App instance of the background thread.

        It is not meant to be used aside from passing to methods using `post`.
        """
        return BackgroundApp.__app

    def _post(self, callable: typing.Callable, try_post: bool = False):
        if BackgroundApp.__stopped:
            raise RuntimeError("Cannot use BackgroundApp after stopping it.")
        if try_post:
            return BackgroundApp.__poster.try_post(callable)
        return BackgroundApp.__poster.post(callable)

    def post(self, callable: typing.Callable):
        """Post callable method to the background app thread."""
        return self._post(callable)

    def try_post(self, callable: typing.Callable):
        """Try post callable method to the background app thread."""
        return self._post(callable, try_post=True)

    def stop(self) -> None:
        """Stop the background app thread."""
        if BackgroundApp.__stopped:
            return
        BackgroundApp.__stop_signaled = True
        while True:
            time.sleep(0.05)
            if BackgroundApp.__stopped:
                break

    def _start_app(self, **kwargs) -> None:
        BackgroundApp.__app = mech.App(**kwargs)
        BackgroundApp.__poster = BackgroundApp.__app.poster
        atexit.register(_exit, self)
        while True:
            if BackgroundApp.__stop_signaled:
                break
            try:
                utils.sleep(40)
            except Exception as e:  # pragma: no cover
                raise Exception("BackgroundApp cannot sleep.") from e  # pragma: no cover
        BackgroundApp.__stopped = True



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Client for Mechanical services."""

from pathlib import Path
import time

import rpyc

from ansys.mechanical.core.embedding.rpc.utils import PYMECHANICAL_DEFAULT_RPC_PORT
from ansys.mechanical.core.mechanical import DEFAULT_CHUNK_SIZE


class Client:
    """Client for connecting to Mechanical services."""

    def __init__(
        self, host: str, port: int, timeout: float = 120.0, cleanup_on_exit=True, process=None
    ):
        """Initialize the client.

        Parameters
        ----------
        host : str, optional
            IP address to connect to the server.  The default is ``None``
            in which case ``localhost`` is used.
        port : int, optional
            Port to connect to the Mecahnical server. The default is ``None``,
            in which case ``20000`` is used.
        timeout : float, optional
            Maximum allowable time for connecting to the Mechanical server.
            The default is ``60.0``.
        process: subprocess.Popen, optional
            The process object that was connected to

        """
        if host is None:
            host = "localhost"
        self.host = host
        if port is None:
            port = PYMECHANICAL_DEFAULT_RPC_PORT
        self.port = port
        self._process = process
        self.timeout = timeout
        self.connection = None
        self.root = None
        self._connect()
        self._cleanup_on_exit = cleanup_on_exit
        self._error_type = Exception
        self._has_exited = False

    def _get_python_script_api_version(self) -> bool:
        """Get the python api version."""
        return 1

    def __getattr__(self, attr):
        """Get attribute from the root object."""
        if hasattr(self.root, attr):
            return getattr(self.root, attr)
        propget_name = f"propget_{attr}"
        if hasattr(self.root, propget_name):
            exposed_fget = getattr(self.root, propget_name)
            return exposed_fget()
        return self.__dict__.items[attr]

    # TODO : Implement setattr
    # def __setattr__(self, attr, value):
    #     if hasattr(self.root, attr):
    #         inner_prop = getattr(self.root.__class__, attr)
    #         if isinstance(inner_prop, property):
    #             inner_prop.fset(self.root, value)
    #     else:
    #         super().__setattr__(attr, value)

    def _connect(self):
        self._wait_until_ready()
        self.root = self.connection.root
        print(f"Connected to {self.host}:{self.port}")

    def _exponential_backoff(self, max_time=60.0, base_time=0.1, factor=2):
        """Generate exponential backoff timing."""
        t_max = time.time() + max_time
        t = base_time
        while time.time() < t_max:
            yield t
            t = min(t * factor, max_time)

    def _wait_until_ready(self):
        """Wait until the server is ready."""
        t_max = time.time() + self.timeout
        for delay in self._exponential_backoff(max_time=self.timeout):
            if time.time() >= t_max:
                break  # Exit if the timeout is reached
            try:
                self.connection = rpyc.connect(self.host, self.port)
                self.connection.ping()
                print("Server is ready.")
                return
            except Exception:
                time.sleep(delay)

        raise TimeoutError(
            f"Server at {self.host}:{self.port} not ready within {self.timeout} seconds."
        )

    def close(self):
        """Close the connection."""
        print("Closing the connection")
        self.connection.close()
        print(f"Connection to {self.host}:{self.port} closed")

    def upload(
        self,
        file_name,
        file_location_destination=None,
        chunk_size=DEFAULT_CHUNK_SIZE,
        progress_bar=False,
    ):
        """Upload a file to the server."""
        print(f"arg: {file_name}, {file_location_destination}")
        print()
        file_path = Path(file_name)
        if not file_path.exists():
            print(f"File {file_name} does not exist.")
            return

        file_base_name = file_path.name
        remote_path = Path(file_location_destination) / file_base_name

        with file_path.open("rb") as f:
            file_data = f.read()
            self.service_upload(str(remote_path), file_data)

        print(f"File {file_name} uploaded to {file_location_destination}")

    def download(
        self,
        files,
        target_dir=None,
        chunk_size=DEFAULT_CHUNK_SIZE,
        progress_bar=None,
        recursive=False,
    ):
        """Download a file from the server."""
        out_files = []
        target_path = Path(target_dir)
        target_path.mkdir(parents=True, exist_ok=True)

        response = self.service_download(files)

        if isinstance(response, dict) and response["is_directory"]:
            files_path = Path(files)
            for relative_file_path in response["files"]:
                full_remote_path = files_path / relative_file_path
                local_file_path = target_path / relative_file_path
                local_file_dir = local_file_path.parent
                local_file_dir.mkdir(parents=True, exist_ok=True)

                out_file_path = self._download_file(
                    str(full_remote_path), str(local_file_path), chunk_size, overwrite=True
                )
        else:
            files_path = Path(files)
            local_file_path = target_path / files_path.name
            out_file_path = self._download_file(
                files, str(local_file_path), chunk_size, overwrite=True
            )
        out_files.append(out_file_path)

        return out_files

    def _download_file(self, remote_file_path, local_file_path, chunk_size=1024, overwrite=False):
        local_path = Path(local_file_path)
        if local_path.exists() and not overwrite:
            print(f"File {local_file_path} already exists locally. Skipping download.")
            return

        response = self.service_download(remote_file_path)
        if isinstance(response, dict):
            raise ValueError("Expected a file download, but got a directory response.")
        file_data = response

        # Write the file data to the local path
        with local_path.open("wb") as f:
            f.write(file_data)

        print(f"File {remote_file_path} downloaded to {local_file_path}")

        return str(local_path)

    def download_project(self, extensions=None, target_dir=None, progress_bar=False):
        """Download all project files in the working directory of the Mechanical instance."""
        destination_directory = target_dir.rstrip("\\/")
        if destination_directory:
            path = Path(destination_directory)
            path.mkdir(parents=True, exist_ok=True)
        else:
            destination_directory = str(Path.cwd())

        # relative directory?
        dest_path = Path(destination_directory)
        if dest_path.is_dir():
            if not dest_path.is_absolute():
                # construct full path
                destination_directory = str(Path.cwd() / destination_directory)

        _project_directory = self.project_directory
        _project_directory = _project_directory.rstrip("\\/")
        project_path = Path(_project_directory)

        # this is where .mechddb resides
        parent_directory = str(project_path.parent)

        list_of_files = []

        if not extensions:
            files = self.list_files()
        else:
            files = []
            for each_extension in extensions:
                # mechdb resides one level above project directory
                if "mechdb" == each_extension.lower():
                    file_temp = str(project_path.parent / f"*.{each_extension}")
                else:
                    file_temp = str(project_path / "**" / f"*.{each_extension}")

                list_files = self._get_files(file_temp, recursive=False)
                files.extend(list_files)

        for file in files:
            # create similar hierarchy locally
            file_path = Path(file)
            parent_path = Path(parent_directory)
            dest_path = Path(destination_directory)

            # Replace parent directory with destination directory
            relative_path = file_path.relative_to(parent_path)
            new_path = dest_path / relative_path
            new_path_dir = new_path.parent

            temp_files = self.download(
                files=file, target_dir=str(new_path_dir), progress_bar=progress_bar
            )
            list_of_files.extend(temp_files)
        return list_of_files

    @property
    def backend(self) -> str:
        """Get the backend type."""
        return "python"

    @property
    def is_alive(self):
        """Check if the Mechanical instance is alive."""
        try:
            self.connection.ping()
            return True
        except Exception as e:
            print(f"Connection check failed: {e}")
            return False

    def exit(self):
        """Shuts down the Mechanical instance."""
        if self._has_exited:
            return
        self.close()
        self._has_exited = True
        print("Disconnected from server")

    def __del__(self):  # pragma: no cover
        """Clean up on exit."""
        if self._cleanup_on_exit:
            try:
                self.exit()
            except Exception as e:
                print(f"Failed to exit cleanly: {e}")



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Mechanical beta feature flags."""

import typing
import warnings


class FeatureFlags:
    """Supported feature flag names."""

    ThermalShells = "Mechanical.ThermalShells"
    MultistageHarmonic = "Mechanical.MultistageHarmonic"
    CPython = "Mechanical.CPython.Capability"


def get_feature_flag_names() -> typing.List[str]:
    """Get the available feature flags."""
    return [x for x in dir(FeatureFlags) if "_" not in x]


def _get_flag_arg(flagname: str) -> str:
    """Get the command line name for a given feature flag."""
    if hasattr(FeatureFlags, flagname):
        return getattr(FeatureFlags, flagname)
    warnings.warn(f"Using undocumented feature flag {flagname}")
    return flagname


def get_command_line_arguments(flags: typing.List[str]):
    """Get the command line arguments as an array for the given flags."""
    return ["-featureflags", ";".join([_get_flag_arg(flag) for flag in flags])]



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""PyVista plotter."""

from __future__ import annotations

import dataclasses
from enum import Enum
import os
import typing
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.mechanical.core.embedding import App
from ansys.tools.visualization_interface import Plotter
import clr
import numpy as np
import pyvista as pv

from .utils import (
    bgr_to_rgb_tuple,
    get_line_nodes_and_coords,
    get_scene_for_object,
    get_tri_nodes_and_coords,
    get_tri_result_disp_and_results,
)

clr.AddReference("Ansys.Mechanical.DataModel")
clr.AddReference("Ansys.ACT.Interfaces")

import Ansys  # noqa: E402


@dataclasses.dataclass
class Plottable:
    """Plottable object."""

    polydata: typing.Optional[pv.PolyData] = None

    # TODO : Make this a list of overridable attributes
    color: typing.Optional[pv.Color] = None
    transform: pv.transform.Transform = None
    children: typing.List["Plottable"] = None
    kwargs: typing.Dict = None

    def __post_init__(self):
        """Initialize the plottable.

        The transform will be identity.
        The children will be an empty list.
        """
        self.transform = np.identity(4)
        self.transform = pv.transform.Transform(np.identity(4))
        self.children = list()


def _transform_to_pyvista(transform: "Ansys.ACT.Math.Matrix4D") -> pv.transform.Transform:
    """Convert the Transformation matrix to a numpy array."""
    np_transform = np.array([transform[i] for i in range(16)]).reshape(4, 4)

    # The mechanical scenegraph transform node is the transpose of the pyvista transform matrix
    np_transform = np_transform.transpose()
    return pv.transform.Transform(np_transform)


def _get_color(node: "Ansys.Mechanical.Scenegraph.AttributeNode") -> pv.Color:
    node_color = node.Property(Ansys.Mechanical.Scenegraph.ScenegraphIntAttributes.Color)
    if node_color is None:
        return None
    color = pv.Color(bgr_to_rgb_tuple(node_color))
    return color


class MeshOrientedTransformResizeStyle(Enum):
    """Dynamic resize style flag for mesh oriented transform nodes."""

    NONE = 0
    SCALING = 1
    STRETCHING = 2


@dataclasses.dataclass
class PlotSettings:
    """Settings for a plot."""

    displacement_scale_factor: float = 1.0
    point_size = 5


class ScenegraphNodeVisitor:
    """Class to visit the Mechanical scenegraph nodes."""

    def __init__(self, app, plot_settings: PlotSettings):
        """Construct a new instance of the visitor class."""
        self._app = app
        self._plot_settings = plot_settings

    def _visit_group_node(self, node: "Ansys.Mechanical.Scenegraph.GroupNode") -> Plottable:
        """Return a new plottable grouping all the children of the group node."""
        plottable = Plottable()
        for child in node.Children:
            child_plottable = self.visit_node(child)
            if child_plottable is not None:
                plottable.children.append(child_plottable)
        return plottable

    def _visit_line_tessellation_node(
        self,
        node: "Ansys.Mechanical.Scenegraph.LineTessellationNode",
    ) -> Plottable:
        np_coordinates, np_indices = get_line_nodes_and_coords(node)
        np_indices = np.insert(np_indices, 0, 2, axis=1)
        line_mesh = pv.PolyData()
        line_mesh.points = np_coordinates
        line_mesh.lines = np_indices
        plottable = Plottable(line_mesh)
        return plottable

    def _visit_tri_tessellation_node(
        self,
        node: "Ansys.Mechanical.Scenegraph.TriTessellationNode",
    ) -> Plottable:
        np_coordinates, np_indices = get_tri_nodes_and_coords(node)
        if np_coordinates is None or np_indices is None:
            return None
        np_indices = np.insert(np_indices, 0, 3, axis=1)
        plottable = Plottable(pv.PolyData(np_coordinates, np_indices))
        return plottable

    def _visit_tri_tessellation_result_node(
        self,
        node: "Ansys.Mechanical.Scenegraph.TriTessellationResultNode",
    ) -> Plottable:
        coords, indices = get_tri_nodes_and_coords(node)
        if coords is None or indices is None:
            return None
        indices = np.insert(indices, 0, 3, axis=1)

        disps, results = get_tri_result_disp_and_results(node)
        deformed_coords = coords + disps

        plottable = Plottable(pv.PolyData(deformed_coords, indices))
        plottable.polydata.point_data["Results"] = results
        plottable.kwargs = {"cmap": "viridis", "show_edges": True, "remove_color": 1}
        return plottable

    def _visit_transform_node(self, node: "Ansys.Mechanical.Scenegraph.TransformNode") -> Plottable:
        plottable = self.visit_node(node.Child)
        if plottable is None:
            return None
        plottable.transform = _transform_to_pyvista(node.Transform)
        return plottable

    def _visit_attribute_node(self, node: "Ansys.Mechanical.Scenegraph.AttributeNode") -> Plottable:
        """Return the plottable of the child node with the color attached."""
        plottable = self.visit_node(node.Child)
        if plottable is None:
            return None
        plottable.color = _get_color(node)
        return plottable

    def _visit_mesh_oriented_transform_node(
        self, node: "Ansys.Mechanical.Scenegraph.MeshOrientedTransformNode"
    ) -> Plottable:
        if "PYMECHANICAL_SCENE_VISIT_MESH_ORIENTED_TRANSFORM_NODE" not in os.environ:
            self._app.log_warning("Ignoring MeshOrientedTransformNode")
            return None

        plottable = self.visit_node(node.Child)
        if plottable is None:
            return None

        xform2 = _transform_to_pyvista(
            node.GetComputedTransform(self._plot_settings.displacement_scale_factor)
        )
        plottable.transform = xform2
        return plottable

    def _visit_point_cloud_node(
        self, node: "Ansys.Mechanical.Scenegraph.PointCloudNod"
    ) -> Plottable:
        point_coords = np.array(node.Coordinates, dtype=np.double)
        point_indices = np.array(node.Indices, dtype=np.int32)
        points = np.zeros(shape=(len(point_indices), 3))
        for loop_index, point_index in enumerate(point_indices):
            point = point_coords[point_index * 3 : point_index * 3 + 3]
            points[loop_index] = point
        plottable = Plottable(pv.PolyData(points))
        plottable.kwargs = {"render_points_as_spheres": True}
        return plottable

    def visit_node(self, node: "Ansys.Mechanical.Scenegraph.Node") -> Plottable:
        """Visit an arbitrary node.

        Return a plottable object of that node.
        """
        if not isinstance(node, Ansys.Mechanical.Scenegraph.Node):
            raise Exception("Node is not a scenegraph node!")

        self._app.log_info(f"Visiting {node}")

        if isinstance(node, Ansys.Mechanical.Scenegraph.GroupNode):
            return self._visit_group_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.AttributeNode):
            return self._visit_attribute_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.TransformNode):
            return self._visit_transform_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.TriTessellationNode):
            return self._visit_tri_tessellation_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.LineTessellationNode):
            return self._visit_line_tessellation_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.TriTessellationResultNode):
            return self._visit_tri_tessellation_result_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.MeshOrientedTransformNode):
            return self._visit_mesh_oriented_transform_node(node)
        elif isinstance(node, Ansys.Mechanical.Scenegraph.PointCloudNode):
            return self._visit_point_cloud_node(node)
        else:
            self._app.log_warning(f"Unexpected node: {node}")
        return None


def _add_plottable(plotter: Plotter, plottable: Plottable, plot_settings: PlotSettings):
    for child in plottable.children:
        child.transform *= plottable.transform
        if child.color is None:
            child.color = plottable.color
        _add_plottable(plotter, child, plot_settings)

    if plottable.polydata is None:
        return

    polydata = plottable.polydata.transform(plottable.transform, inplace=True)
    kwargs = {
        "color": plottable.color,
        "smooth_shading": True,
        "point_size": plot_settings.point_size,
    }
    if plottable.kwargs is not None:
        kwargs.update(plottable.kwargs)
    if kwargs.get("remove_color", None):
        kwargs.pop("remove_color")
        if "color" in kwargs:
            kwargs.pop("color")
    plotter.plot(polydata, **kwargs)


def _get_plotter_for_scene(
    app: App,
    node: "Ansys.Mechanical.Scenegraph.Node",
    plot_settings: PlotSettings,
) -> Plotter:
    visitor = ScenegraphNodeVisitor(app, plot_settings)
    plottable = visitor.visit_node(node)
    plotter = Plotter()
    _add_plottable(plotter, plottable, plot_settings)
    return plotter


def _plot_object(app: App, obj, plot_settings: PlotSettings) -> Plotter:
    """Get a ``ansys.tools.visualization_interface.Plotter`` instance for `obj`."""
    scene = get_scene_for_object(app, obj)
    if scene is None:
        app.log_warning(f"No scene available for object {obj}!")
        return None
    plotter = _get_plotter_for_scene(app, scene, plot_settings)
    return plotter


def to_plotter(app: App, obj=None, plot_settings: PlotSettings = None) -> Plotter:
    """Convert the scene for `obj` to an ``ansys.tools.visualization_interface.Plotter`` instance.

    If the `obj` is None, default to the Geometry object.
    Geometry, Mesh, and some Results are currently supported.
    """
    if obj is None:
        obj = app.Model.Geometry
    if plot_settings is None:
        plot_settings = PlotSettings()
    return _plot_object(app, obj, plot_settings)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Initializer for Mechanical embedding. Sets up paths and resolvers."""

from importlib.metadata import distribution
import os
from pathlib import Path
import platform
import sys
import warnings

from ansys.mechanical.core.embedding.loader import load_clr
from ansys.mechanical.core.embedding.resolver import resolve
from ansys.mechanical.core.mechanical import LOG

INITIALIZED_VERSION = None
"""Constant for the initialized version."""

SUPPORTED_MECHANICAL_EMBEDDING_VERSIONS = {
    252: "2025R2",
    251: "2025R1",
    242: "2024R2",
    241: "2024R1",
}
"""Supported Mechanical embedding versions on Windows."""


def __add_sys_path(version: int) -> str:
    install_path = Path(os.environ[f"AWP_ROOT{version}"])
    platform_string = "winx64" if os.name == "nt" else "linx64"
    bin_path = install_path / "aisol" / "bin" / platform_string
    sys.path.append(str(bin_path.resolve()))


def __workaround_material_server(version: int) -> None:
    """Workaround material server bug in 2024 R1.

    A REST server is used as a backend for the material model GUI.
    In 2024 R1, this GUI is used even in batch mode. The server
    starts by default on a background thread, which may lead to
    a race condition on shutdown.
    """
    if version in [241]:
        os.environ["ENGRDATA_SERVER_SERIAL"] = "1"


def __check_for_supported_version(version) -> None:
    """Check if Mechanical version is supported with current version of PyMechanical.

    If specific environment variable is enabled, then users can overwrite the supported versions.
    However, using unsupported versions may cause issues.
    """
    allow_old_version = os.getenv("ANSYS_MECHANICAL_EMBEDDING_SUPPORT_OLD_VERSIONS") == "1"

    # Check if the version is supported
    if not allow_old_version and version < min(SUPPORTED_MECHANICAL_EMBEDDING_VERSIONS):
        raise ValueError(f"Mechanical version {version} is not supported.")

    return version


def _get_latest_default_version() -> int:
    """Try to get the latest Mechanical version from the environment.

    Checks if multiple versions of Mechanical found in system.
    For Linux it will be only one since ``mechanical-env`` takes care of that.
    If multiple versions are detected, select the latest one, as no specific version is provided.
    """
    awp_roots = [value for key, value in os.environ.items() if key.startswith("AWP_ROOT")]
    if not awp_roots:
        raise Exception("No Mechanical installations found.")

    versions_found = []
    for path in awp_roots:
        resolved_path = Path(path).resolve()
        folder = resolved_path.name
        if folder.startswith("v") and folder[1:].isdigit():
            version = int(folder[1:])
            versions_found.append(version)

    LOG.info(f"Available versions of Mechanical: {versions_found}")

    latest_version = max(versions_found)

    if len(awp_roots) > 1:
        LOG.warning(
            f"Multiple versions of Mechanical found! Using latest version {latest_version} ..."
        )
    return latest_version


def __check_python_interpreter_architecture() -> None:
    """Embedding support only 64 bit architecture."""
    if platform.architecture()[0] != "64bit":
        raise Exception("Mechanical Embedding requires a 64-bit Python environment.")


def __windows_store_workaround(version: int) -> None:
    """Workaround for Windows store.

    See https://github.com/ansys/pymechanical/issues/1136

    Windows store Python uses the Win32 API SetDefaultDllDirectories
    so that the PATH environment variable isn't scanned for any DLL
    dependency.

    PyMechanical loads the embedding library which automatically sets
    these Paths, but this uses the PATH environment variable which doesn't
    work for Windows store Python.

    We provide a workaround for versions 2024R2 and 2025R1 that sets
    these paths using `os.add_dll_directory`.

    Note:   This workaround does not include DLL paths used in FSI
            mapping workflows.

    Parameters
    ----------
    version : int
        The version of Mechanical to set the DLL paths for.
    """
    # Nothing to do on Linux
    if os.name != "nt":
        return

    # Nothing to do if it isn't a Windows store application
    if r"WindowsApps\PythonSoftwareFoundation" not in sys.base_prefix:
        return

    # Get the AWP_ROOT environment variable for the specified version
    awp_root = Path(os.environ[f"AWP_ROOT{version}"])
    # Set paths to the aisol and framework DLLs
    paths = [
        awp_root / "aisol" / "bin" / "winx64",
        awp_root / "Framework" / "bin" / "Win64",
    ]
    # Set the path to the tp directory within the AWP_ROOTXYZ directory
    awp_root_tp = awp_root / "tp"
    # Add paths to the IntelCompiler, IntelMKL, HDF5, and Qt DLLs for 2024R2 and 2025R1
    if version == 242:
        paths.extend(
            [
                awp_root_tp / "IntelCompiler" / "2023.1.0" / "winx64",
                awp_root_tp / "IntelMKL" / "2023.1.0" / "winx64",
                awp_root_tp / "hdf5" / "1.12.2" / "winx64",
                awp_root_tp / "qt" / "5.15.16" / "winx64" / "bin",
            ]
        )
    elif version == 251:
        paths.extend(
            [
                awp_root_tp / "IntelCompiler" / "2023.1.0" / "winx64",
                awp_root_tp / "IntelMKL" / "2023.1.0" / "winx64",
                awp_root_tp / "hdf5" / "1.12.2" / "winx64",
                awp_root_tp / "qt" / "5.15.17" / "winx64" / "bin",
            ]
        )
    elif version == 252:
        paths.extend(
            [
                awp_root_tp / "IntelCompiler" / "2023.1.0" / "winx64",
                awp_root_tp / "IntelMKL" / "2024.2.3" / "winx64",
                awp_root_tp / "hdf5" / "winx64",
                awp_root_tp / "qt" / "5.15.18" / "winx64" / "bin",
            ]
        )
    else:
        return

    # Add each path to the DLL search path
    for path in paths:
        os.add_dll_directory(str(path))


def __set_environment(version: int) -> None:
    """Set environment variables to configure embedding."""
    if os.name == "nt":  # pragma: no cover
        if version < 251:
            os.environ["MECHANICAL_STARTUP_UNOPTIMIZED"] = "1"

    # Set an environment variable to use the custom CLR host
    # for embedding.
    # In the future (>252), it would always be used.
    if version == 251 or version == 252:
        if "PYMECHANICAL_NO_CLR_HOST_LITE" not in os.environ:
            os.environ["ANSYS_MECHANICAL_EMBEDDING_CLR_HOST"] = "1"
    if version > 252:
        if "PYMECHANICAL_NO_CLR_HOST_LITE" in os.environ:
            os.environ["ANSYS_MECHANICAL_EMBEDDING_CLR_HOST"] = "0"


def __check_for_mechanical_env():
    """Embedding in linux platform must use mechanical-env."""
    if platform.system() == "Linux" and os.environ.get("PYMECHANICAL_EMBEDDING") != "TRUE":
        raise Exception(
            "On linux, embedding an instance of the Mechanical process using"
            "the App class requires running python inside of a Mechanical environment."
            "Use the `mechanical-env` script to do this. For more information, refer to:"
            "https://mechanical.docs.pyansys.com/version/stable/"
            "getting_started/running_mechanical.html#embed-a-mechanical-instance"
        )


def __is_lib_loaded(libname: str):  # pragma: no cover
    """Return whether a library is loaded."""
    import ctypes

    rtld_noload = 4
    try:
        ctypes.CDLL(libname, rtld_noload)
    except OSError:
        return False
    return True


def __check_loaded_libs(version: int = None):  # pragma: no cover
    """Ensure that incompatible libraries aren't loaded prior to PyMechanical load."""
    if platform.system() != "Linux":
        return

    if version < 251:
        return

    # For 2025 R1, PyMechanical will crash on shutdown if libX11.so is already loaded
    # before starting Mechanical
    if __is_lib_loaded("libX11.so"):
        LOG.warning(
            "libX11.so is loaded prior to initializing the Embedded Instance of Mechanical.\
                      Python will crash on shutdown..."
        )


def initialize(version: int = None):
    """Initialize Mechanical embedding."""
    global INITIALIZED_VERSION
    if version is None:
        version = _get_latest_default_version()

    version = __check_for_supported_version(version=version)

    if INITIALIZED_VERSION is not None:
        if INITIALIZED_VERSION != version:
            raise ValueError(
                f"Initialized version {INITIALIZED_VERSION} "
                f"does not match the expected version {version}."
            )
        return INITIALIZED_VERSION

    __check_python_interpreter_architecture()  # blocks 32 bit python
    __check_for_mechanical_env()  # checks for mechanical-env in linux embedding

    __set_environment(version)

    __windows_store_workaround(version)

    __check_loaded_libs(version)

    __workaround_material_server(version)

    # need to add system path in order to import the assembly with the resolver
    __add_sys_path(version)

    # Check if 'pythonnet' is installed... and if so, throw warning
    try:
        distribution("pythonnet")
        warnings.warn(
            "The pythonnet package was found in your environment "
            "which interferes with the ansys-pythonnet package. "
            "Some APIs may not work due to pythonnet being installed.\n\n"
            "For using PyMechanical, we recommend you do the following:\n"
            "1. Uninstall your existing pythonnet package: pip uninstall pythonnet\n"
            "2. Install the ansys-pythonnet package: pip install --upgrade "
            "--force-reinstall ansys-pythonnet\n",
            stacklevel=2,
        )
    except ModuleNotFoundError:
        pass

    # load the CLR with mono that is shipped with the unified ansys installer
    load_clr(os.environ[f"AWP_ROOT{version}"])

    # attach the resolver
    resolve(version)

    INITIALIZED_VERSION = version
    return version



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Internal Mechanical logging Linux API.

.. note::
   The Linux API is needed for version 2023 R2, where the .NET backend is Windows-only.
   This API does not support some options, namely the base directory and log filename.

"""

import ctypes
import logging
import os
from pathlib import Path

from ansys.mechanical.core.embedding import initializer
from ansys.mechanical.core.embedding.logger import sinks


def _get_dll():
    installdir = os.environ[f"AWP_ROOT{initializer.INITIALIZED_VERSION}"]
    dll = ctypes.CDLL(str(Path(installdir) / "aisol/dll/linx64/libAns.Common.WBLogger.so"))
    dll.wb_logger_enable_sink.argtypes = [ctypes.c_int32]

    dll.wb_logger_disable_sink.argtypes = [ctypes.c_int32]

    dll.wb_logger_set_sink_filter_level.argtypes = [ctypes.c_int32, ctypes.c_int32]

    dll.wb_logger_flush.argtypes = []

    dll.wb_logger_set_logging_auto_flush.argtypes = [ctypes.c_int8]

    dll.wb_logger_set_logging_directory.argtypes = [ctypes.c_char_p]

    dll.wb_logger_can_log_message.argtypes = [ctypes.c_int32]
    dll.wb_logger_can_log_message.restype = ctypes.c_int32

    dll.wb_logger_log_message.argtypes = [ctypes.c_int32, ctypes.c_char_p, ctypes.c_char_p]
    return dll


def _get_sink_id(standard_sink_type: int) -> ctypes.c_int32:
    """Convert standard sink type to sink id."""
    return {
        sinks.StandardSinks.STANDARD_LOG_FILE: ctypes.c_int32(1),
        sinks.StandardSinks.CONSOLE: ctypes.c_int32(2),
    }[standard_sink_type]


def _str_to_utf8_ptr(value: str) -> ctypes.c_char_p:
    return ctypes.create_string_buffer(value.encode())


def _bool_to_single_byte_int(value: bool) -> ctypes.c_int8:
    if value:
        return ctypes.c_int8(1)
    return ctypes.c_int8(0)


def _to_wb_logger_severity(level: int) -> ctypes.c_int32:
    """Convert to internal integer."""
    if level <= logging.DEBUG:
        # Level 0 in Workbench logging is trace. Level 1 is debug.
        # Python logging.DEBUG implies trace.
        return ctypes.c_int32(0)
    elif level <= logging.INFO:
        return ctypes.c_int32(2)
    elif level <= logging.WARNING:
        return ctypes.c_int32(3)
    elif level <= logging.ERROR:
        return ctypes.c_int32(4)
    elif level <= logging.CRITICAL:
        return ctypes.c_int32(5)
    else:
        # default to info
        return ctypes.c_int32(2)


class APIBackend:
    """Provides the API backend for the Mechanical logging system."""

    def flush(self) -> None:
        """Flush the log manually."""
        return _get_dll().wb_logger_flush()

    def enable(self, sink: int = sinks.StandardSinks.CONSOLE) -> None:
        """Enable logging."""
        sinkid = _get_sink_id(sink)
        _get_dll().wb_logger_enable_sink(sinkid)

    def disable(self, sink: int = sinks.StandardSinks.CONSOLE) -> None:
        """Disable logging."""
        sinkid = _get_sink_id(sink)
        _get_dll().wb_logger_disable_sink(sinkid)

    def set_log_level(self, level: int, sink: int = sinks.StandardSinks.CONSOLE) -> None:
        """Set the log level for Mechanical based on the Python log level."""
        if level == logging.NOTSET:
            self.disable(sink)

        sinkid = _get_sink_id(sink)
        wb_level = _to_wb_logger_severity(level)
        _get_dll().wb_logger_set_sink_filter_level(sinkid, wb_level)

    def set_auto_flush(self, flush: bool) -> None:
        """Set whether to auto flush to the standard log file."""
        flag = _bool_to_single_byte_int(flush)
        _get_dll().wb_logger_set_logging_auto_flush(flag)

    def set_directory(self, directory: str) -> None:
        """Set the location to write the log file to."""
        value = _str_to_utf8_ptr(directory)
        _get_dll().wb_logger_set_logging_directory(value)

    def set_base_directory(self, base_directory: str) -> None:
        """Set the base location to write the log file to.

        The base directory contains time-stamped subfolders where the log file
        is actually written to. If a base directory is set, it takes precedence over the
        ``set_directory`` location.

        This does not have an API to set at runtime.
        """
        raise Exception("Base directory can only be set before starting the Mechanical instance.")

    def can_log_message(self, level: int) -> bool:
        """Return whether a message with the given severity is outputted to the log."""
        wb_level = _to_wb_logger_severity(level)
        return bool(_get_dll().wb_logger_can_log_message(wb_level))

    def log_message(self, level: int, context: str, message: str) -> None:
        """Log the message to the configured logging mechanism."""
        wb_level = _to_wb_logger_severity(level)
        wb_context = _str_to_utf8_ptr(context)
        wb_message = _str_to_utf8_ptr(message)
        _get_dll().wb_logger_log_message(wb_level, wb_context, wb_message)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Remote Procedure Call (RPC) server."""

import fnmatch
from pathlib import Path

from ansys.mechanical.core.embedding.app import App

from .server import MechanicalEmbeddedServer
from .utils import remote_method


class DefaultServiceMethods:
    """Default service methods for MechanicalEmbeddedServer."""

    def __init__(self, app: App):
        """Initialize the DefaultServiceMethods."""
        self._app = app

    def __repr__(self):
        """Return the representation of the instance."""
        return '"ServiceMethods instance"'

    @remote_method
    def run_python_script(
        self, script: str, enable_logging=False, log_level="WARNING", progress_interval=2000
    ):
        """Run scripts using Internal python engine."""
        result = self._app.execute_script(script)
        return result

    @remote_method
    def run_python_script_from_file(
        self,
        file_path: str,
        enable_logging=False,
        log_level="WARNING",
        progress_interval=2000,
    ):
        """Run scripts using Internal python engine."""
        return self._app.execute_script_from_file(file_path)

    @remote_method
    def clear(self):
        """Clear the current project."""
        self._app.new()

    @property
    @remote_method
    def project_directory(self):
        """Get the project directory."""
        return self._app.ExtAPI.DataModel.Project.ProjectDirectory

    @remote_method
    def list_files(self):
        """List all files in the project directory."""
        file_list = []
        mechdb_path = self._app.ExtAPI.DataModel.Project.FilePath
        if mechdb_path != "":
            file_list.append(mechdb_path)
        root_dir = self._app.ExtAPI.DataModel.Project.ProjectDirectory

        root_path = Path(root_dir)
        for file_path in root_path.rglob("*"):
            if file_path.is_file():
                file_list.append(str(file_path))

        files_out = "\n".join(file_list).splitlines()
        if not files_out:  # pragma: no cover
            print("No files listed")
        return files_out

    @remote_method
    def _get_files(self, files, recursive=False):
        self_files = self.list_files()  # to avoid calling it too much

        if isinstance(files, str):
            if files in self_files:
                list_files = [files]
            elif "*" in files:
                list_files = fnmatch.filter(self_files, files)
                if not list_files:
                    raise ValueError(
                        f"The `'files'` parameter ({files}) didn't match any file using "
                        f"glob expressions in the remote server."
                    )
            else:
                raise ValueError(
                    f"The `'files'` parameter ('{files}') does not match any file or pattern."
                )

        elif isinstance(files, (list, tuple)):
            if not all([isinstance(each, str) for each in files]):
                raise ValueError(
                    "The parameter `'files'` can be a list or tuple, but it "
                    "should only contain strings."
                )
            list_files = files
        else:
            raise ValueError(
                f"The `file` parameter type ({type(files)}) is not supported."
                "Only strings, tuple of strings, or list of strings are allowed."
            )

        return list_files


class MechanicalDefaultServer(MechanicalEmbeddedServer):
    """Default server with default service methods."""

    def __init__(self, **kwargs):
        """Initialize the MechanicalDefaultServer."""
        super().__init__(impl=DefaultServiceMethods, **kwargs)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Additional imports for embedded Mechanical."""

from __future__ import annotations

import typing
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ansys.mechanical.core.embedding import App


def global_entry_points(app: App) -> typing.Dict:
    """Return the global entry points of the application."""
    vars = {}
    vars["ExtAPI"] = app.ExtAPI
    vars["DataModel"] = app.DataModel
    vars["Model"] = app.DataModel.Project.Model
    vars["Tree"] = app.DataModel.Tree
    vars["Graphics"] = app.ExtAPI.Graphics
    return vars


def global_variables(app: App, enums: bool = False) -> typing.Dict:
    """Return the Mechanical scripting global variables as a dict.

    It can be used to add all of these as global variables in python
    with this command:

    ``globals().update(global_variables(embedded_app))``

    To also import all the enums, set the parameter enums to true.
    """
    vars = global_entry_points(app)

    from ansys.mechanical.core.embedding.app import is_initialized
    from ansys.mechanical.core.embedding.transaction import Transaction

    vars["Transaction"] = Transaction

    # Import modules if the app is initialized
    if is_initialized():
        from ansys.mechanical.core.embedding.global_importer import (
            Ansys,
            MechanicalEnums,
            Point,
            Point2D,
            Point3D,
            Quantity,
            SectionPlane,
            System,
            Vector3D,
        )

        vars["Quantity"] = Quantity
        vars["System"] = System
        vars["Ansys"] = Ansys
        vars["MechanicalEnums"] = MechanicalEnums
        # Graphics
        vars["Point"] = Point
        vars["SectionPlane"] = SectionPlane
        # Math
        vars["Point2D"] = Point2D
        vars["Point3D"] = Point3D
        vars["Vector3D"] = Vector3D

    if enums:
        vars.update(get_all_enums())

    return vars


def get_all_enums() -> typing.Dict[str, typing.Any]:
    """Get all the enums as a dictionary."""
    import ansys.mechanical.core.embedding.enum_importer as enum_importer

    enums = {}
    for attr in dir(enum_importer):
        if not hasattr(enum_importer, attr):
            continue
        the_enum = getattr(enum_importer, attr)
        if type(the_enum).__name__ == "CLRMetatype":
            enums[attr] = the_enum
    return enums



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Library to import Mechanical enums.

A useful subset of what is imported by
Ansys Inc/v{NNN}/ACT/apis/Mechanical.py
"""

from ansys.mechanical.core.embedding.app import is_initialized

if not is_initialized():
    raise Exception("Enums cannot be imported until the embedded app is initialized.")

import clr

clr.AddReference("Ansys.Mechanical.DataModel")
clr.AddReference("Ansys.ACT.Interfaces")

from Ansys.ACT.Interfaces.Common import *  # noqa isort: skip
from Ansys.Mechanical.DataModel.Enums import *  # noqa isort: skip
from Ansys.ACT.Interfaces.Analysis import *  # noqa isort: skip
from Ansys.ACT.Interfaces.Geometry import *  # noqa isort: skip
from Ansys.ACT.Mechanical.Fields import *  # noqa isort: skip
from Ansys.ACT.Interfaces.Mesh import *  # noqa isort: skip
import Ansys  # noqa  isort: skip



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Convenience CLI to run mechanical."""

import json
import os
from pathlib import Path
import re
import site
import sys
import warnings

import ansys.tools.path as atp
import click


def get_stubs_location():
    """Find the ansys-mechanical-stubs installation location in site-packages.

    Returns
    -------
    pathlib.Path
        The path to the ansys-mechanical-stubs installation in site-packages.
    """
    site_packages = site.getsitepackages()
    prefix_path = sys.prefix.replace("\\", "\\\\")
    site_packages_regex = re.compile(f"{prefix_path}.*site-packages$")
    site_packages_paths = list(filter(site_packages_regex.match, site_packages))

    if len(site_packages_paths) >= 1:
        # Get the stubs location
        stubs_location = Path(site_packages_paths[0]) / "ansys" / "mechanical" / "stubs"
        return stubs_location

    raise Exception("Could not retrieve the location of the ansys-mechanical-stubs package.")


def get_stubs_versions(stubs_location: Path):
    """Retrieve the revision numbers in ansys-mechanical-stubs.

    Parameters
    ----------
    pathlib.Path
        The path to the ansys-mechanical-stubs installation in site-packages.

    Returns
    -------
    list
        The list containing minimum and maximum versions in the ansys-mechanical-stubs package.
    """
    # Get revision numbers in stubs folder
    revns = [
        int(revision.name[1:])
        for revision in stubs_location.iterdir()
        if revision.is_dir() and revision.name.startswith("v")
    ]
    return revns


def _vscode_impl(
    target: str = "user",
    revision: int = None,
):
    """Get the IDE configuration for autocomplete in VS Code.

    Parameters
    ----------
    target: str
        The type of settings to update. Either "user" or "workspace" in VS Code.
        By default, it's ``user``.
    revision: int
        The Mechanical revision number. For example, "252".
        If unspecified, it finds the default Mechanical version from ansys-tools-path.
    """
    # Update the user or workspace settings
    settings_json = "the settings.json file"
    if target == "user":
        # Get the path to the user's settings.json file depending on the platform
        if "win" in sys.platform:
            settings_json = (
                Path(os.environ.get("APPDATA")) / "Code" / "User" / "settings.json"
            )  # pragma: no cover
        elif "lin" in sys.platform:
            settings_json = (
                Path(os.environ.get("HOME")) / ".config" / "Code" / "User" / "settings.json"
            )
    elif target == "workspace":
        # Get the current working directory
        current_dir = Path.cwd()
        # Get the path to the settings.json file based on the git root & .vscode folder
        settings_json = current_dir / ".vscode" / "settings.json"

    # Location where the stubs are installed -> .venv/Lib/site-packages, for example
    stubs_location = get_stubs_location() / f"v{revision}"

    # The settings to add to settings.json for autocomplete to work
    settings_json_data = {
        "python.autoComplete.extraPaths": [str(stubs_location)],
        "python.analysis.extraPaths": [str(stubs_location)],
    }
    # Pretty print dictionary
    pretty_dict = json.dumps(settings_json_data, indent=4)

    print(f"Update {settings_json} with the following information:\n")

    if target == "workspace":
        print(
            "Note: Please ensure the .vscode folder is in the root of your project or repository.\n"
        )

    print(pretty_dict)


def _cli_impl(
    ide: str = "vscode",
    target: str = "user",
    revision: int = None,
):
    """Provide the user with the path to the settings.json file and IDE settings.

    Parameters
    ----------
    ide: str
        The IDE to set up autocomplete settings. By default, it's ``vscode``.
    target: str
        The type of settings to update. Either "user" or "workspace" in VS Code.
        By default, it's ``user``.
    revision: int
        The Mechanical revision number. For example, "252".
        If unspecified, it finds the default Mechanical version from ansys-tools-path.
    """
    # Get the ansys-mechanical-stubs install location
    stubs_location = get_stubs_location()
    # Get all revision numbers available in ansys-mechanical-stubs
    revns = get_stubs_versions(stubs_location)

    # Check if the user revision number is less or greater than the min and max revisions
    # in the ansys-mechanical-stubs package location
    if revision < min(revns):
        raise Exception(f"PyMechanical Stubs are not available for {revision}")
    elif revision > max(revns):
        warnings.warn(
            f"PyMechanical Stubs are not available for {revision}. Using {max(revns)} instead.",
            stacklevel=2,
        )
        revision = max(revns)

    # Check the IDE and raise an exception if it's not VS Code
    if ide != "vscode":
        raise Exception(f"{ide} is not supported at the moment.")

    return _vscode_impl(target, revision)


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "--ide",
    default="vscode",
    type=str,
    help="The IDE being used. By default, it's ``vscode``.",
)
@click.option(
    "--target",
    default="user",
    type=str,
    help="The type of settings to update - either ``user`` or ``workspace`` settings in VS Code.",
)
@click.option(
    "--revision",
    default=None,
    type=int,
    help='The Mechanical revision number, e.g. "252" or "251". If unspecified,\
it finds and uses the default version from ansys-tools-path.',
)
def cli(ide: str, target: str, revision: int) -> None:
    """CLI tool to update settings.json files for autocomplete with ansys-mechanical-stubs.

    Parameters
    ----------
    ide: str
        The IDE to set up autocomplete settings. By default, it's ``vscode``.
    target: str
        The type of settings to update. Either "user" or "workspace" in VS Code.
        By default, it's ``user``.
    revision: int
        The Mechanical revision number. For example, "252".
        If unspecified, it finds the default Mechanical version from ansys-tools-path.

    Usage
    -----
    The following example demonstrates the main use of this tool:

        $ ansys-mechanical-ideconfig --ide vscode --target user --revision 252

    """
    if not revision:
        exe = atp.get_mechanical_path(allow_input=False, version=revision)
        version = atp.version_from_path("mechanical", exe)
    else:
        version = revision

    return _cli_impl(
        ide,
        target,
        version,
    )



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Common plotting utilities."""

from __future__ import annotations

import os
import typing
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from ansys.mechanical.core.embedding import App


def bgr_to_rgb_tuple(bgr_int: int) -> typing.Tuple[int, int, int]:
    """Convert bgr integer to rgb tuple."""
    r = bgr_int & 255
    g = (bgr_int >> 8) & 255
    b = (bgr_int >> 16) & 255
    return r, g, b


def _reshape_ncols(arr: np.array, ncols: int, name: str = "array"):
    """Reshapes the given array into `ncols` columns.

    Precondition - the array's length must be divisible by `ncols`.
    """
    err = f"{name} must be of the form (x0,y0,z0,x1,y1,z1,...,xn,yn,zn).\
        Given {name} are not divisible by 3!"
    if arr.size % ncols != 0:
        raise ValueError(err)
    numrows = int(arr.size / ncols)
    arr = np.reshape(arr, (numrows, ncols))
    return arr


def get_line_nodes_and_coords(
    line_tessellation: "Ansys.Mechanical.Scenegraph.LineTessellationNode",  # noqa: F821
):
    """Extract the nodes and coordinates from the LineTessellationNode.

    The TriTessellationNode contains "Coordinates" and "Indices"
    that are flat arrays. This function converts them to numpy arrays
    """
    np_coordinates = _reshape_ncols(
        np.array(line_tessellation.Coordinates, dtype=np.double), 3, "coordinates"
    )

    np_indices = _reshape_ncols(np.array(line_tessellation.Indices, dtype=np.int32), 2, "indices")
    return np_coordinates, np_indices


def get_tri_nodes_and_coords(tri_tessellation: "Ansys.Mechanical.Scenegraph.TriTessellationNode"):  # noqa: F821
    """Extract the nodes and coordinates from the TriTessellationNode.

    The TriTessellationNode contains "Coordinates" and "Indices"
    that are flat arrays. This function converts them to numpy arrays of the appropriate shape.
    """
    np_coordinates = _reshape_ncols(
        np.array(tri_tessellation.Coordinates, dtype=np.double), 3, "coordinates"
    )
    np_indices = _reshape_ncols(np.array(tri_tessellation.Indices, dtype=np.int32), 3, "indices")
    return np_coordinates, np_indices


def get_tri_result_disp_and_results(
    tri_tessellation: "Ansys.Mechanical.Scenegraph.TriTessellationResultNode",  # noqa: F821
):
    """Extract the defomation and results from the TriTessellationResultNode.

    The TriTessellationResultNode contains "Displacements" and "Results"
    that are flat arrays. This function converts them to numpy arrays of the appropriate shape.
    """
    np_disp = _reshape_ncols(
        np.array(tri_tessellation.Displacements, dtype=np.double), 3, "deformation"
    )
    np_results = np.array(tri_tessellation.Results, dtype=np.double)
    return np_disp, np_results


def _get_geometry_scene(
    app: App,
) -> "Ansys.Mechanical.Scenegraph.GroupNode":  # noqa: F821
    """Get the scene for the geometry.

    using the undocumented apis under ScenegraphHelpers.
    """
    import clr

    clr.AddReference("Ansys.Mechanical.DataModel")
    clr.AddReference("Ansys.Mechanical.Scenegraph")
    clr.AddReference("Ansys.ACT.Interfaces")

    import Ansys  # isort: skip

    category = Ansys.Mechanical.DataModel.Enums.DataModelObjectCategory.Body
    group_node = Ansys.Mechanical.Scenegraph.Builders.GroupNodeBuilder()
    for body in app.DataModel.GetObjectsByType(category):
        scenegraph_node = Ansys.ACT.Mechanical.Tools.ScenegraphHelpers.GetScenegraph(body)
        # wrap the body node in an attribute node using the body color
        attribute_node_builder = Ansys.Mechanical.Scenegraph.Builders.AttributeNodeBuilder()
        attribute_node = (
            attribute_node_builder.Tag(f"body{body.ObjectId}")
            .Child(scenegraph_node)
            # set the color, body.Color is a BGR uint bitfield
            .Property(Ansys.Mechanical.Scenegraph.ScenegraphIntAttributes.Color, body.Color)
            .Build()
        )
        group_node.AddChild(attribute_node)
    return group_node.Build()


def get_scene(
    app: App,
) -> "Ansys.Mechanical.Scenegraph.GroupNode":  # noqa: F821
    """Get the scene of the model."""
    return _get_geometry_scene(app)


def _get_scene_for_object(app: App, obj) -> "Ansys.Mechanical.Scenegraph.Node":  # noqa: F821
    from Ansys.Mechanical.DataModel.Enums import DataModelObjectCategory

    if obj.DataModelObjectCategory == DataModelObjectCategory.Geometry:
        scene = get_scene(app)
        return scene
    if app.version < 261:
        return None
    active_objects = app.Tree.ActiveObjects
    app.Tree.Activate([obj])
    scenegraph_node = app.Graphics.GetScenegraphForActiveObject()
    app.Tree.Activate(active_objects)
    return scenegraph_node


def get_scene_for_object(app: App, obj) -> "Ansys.Mechanical.Scenegraph.Node":  # noqa: F821
    """Get the scene for the given object.

    2025R2 and before: only geometry is supported
    later, Mesh and some Results will be supported.
    """
    node = _get_scene_for_object(app, obj)
    if node is not None:
        save_file = os.environ.get("PYMECHANICAL_SAVE_SCENE_FILE", None)
        if save_file is not None:
            import Ansys

            Ansys.Mechanical.Scenegraph.Persistence.SaveToFile(save_file, node)
    return node



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Utilities necessary for remote calls."""

import typing

from ansys.mechanical.core.mechanical import port_in_use

PYMECHANICAL_DEFAULT_RPC_PORT = 20000


class remote_method:  # noqa: N801
    """Decorator for passing remote methods."""

    def __init__(self, func):
        """Initialize with the given function."""
        self._func = func

    def __call__(self, *args, **kwargs):
        """Call the stored function with provided arguments."""
        return self._func(*args, **kwargs)

    def __call_method__(self, instance, *args, **kwargs):
        """Call the stored function with the instance and provided arguments."""
        return self._func(instance, *args, **kwargs)

    def __get__(self, obj, objtype):
        """Return a partially applied method."""
        from functools import partial

        func = partial(self.__call_method__, obj)
        func._is_remote = True
        func.__name__ = self._func.__name__
        func._owner = obj
        return func


class MethodType:
    """Enum for method or property types."""

    METHOD = 0
    PROP = 1


def try_get_remote_method(methodname: str, obj: typing.Any) -> typing.Tuple[str, typing.Callable]:
    """Try to get a remote method."""
    method = getattr(obj, methodname)
    if not callable(method):
        return None
    if hasattr(method, "_is_remote") and method._is_remote is True:
        return (methodname, method)


def try_get_remote_property(attrname: str, obj: typing.Any) -> typing.Tuple[str, property]:
    """Try to get a remote property."""
    objclass: typing.Type = obj.__class__
    class_attribute = getattr(objclass, attrname)
    getmethod = None
    setmethod = None

    if class_attribute.fget:
        if isinstance(class_attribute.fget, remote_method):
            getmethod = class_attribute.fget
            getmethod._owner = obj
    if class_attribute.fset:
        if isinstance(class_attribute.fset, remote_method):
            setmethod = class_attribute.fset
            setmethod._owner = obj

    return (attrname, property(getmethod, setmethod))


def get_remote_methods(
    obj,
) -> typing.Generator[typing.Tuple[str, typing.Callable, MethodType], None, None]:
    """Yield names and methods of an object's remote methods.

    A remote method is identified by the presence of an attribute `_is_remote` set to `True`.

    Parameters
    ----------
    obj: Any
        The object to inspect for remote methods.

    Yields
    ------
    Generator[Tuple[str, Callable], None, None]
        A tuple containing the method name and the method itself
        for each remote method found in the object
    """
    objclass = obj.__class__
    for attrname in dir(obj):
        if attrname.startswith("__"):
            continue
        if hasattr(objclass, attrname):
            class_attribute = getattr(objclass, attrname)
            if isinstance(class_attribute, property):
                attrname, prop = try_get_remote_property(attrname, obj)
                yield attrname, prop, MethodType.PROP
                continue
        result = try_get_remote_method(attrname, obj)
        if result is not None:
            attrname, method = result
            yield attrname, method, MethodType.METHOD


def get_free_port(port: int = None):
    """Get free port.

    If port is not given, it will find a free port starting from PYMECHANICAL_DEFAULT_RPC_PORT.
    """
    if port is None:
        port = PYMECHANICAL_DEFAULT_RPC_PORT

    while port_in_use(port):
        port += 1

    return port



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""A class to speed up bulk user interactions using Ansys ACT Mechanical Transaction."""


class Transaction:  # When ansys-pythonnet issue #14 is fixed, this class will be removed
    """
    A class to speed up bulk user interactions using Ansys ACT Mechanical Transaction.

    Example
    -------
    >>> with Transaction() as transaction:
    ...     pass  # Perform bulk user interactions here
    """

    def __init__(self):
        """Initialize the Transaction class."""
        import clr

        clr.AddReference("Ansys.ACT.WB1")
        import Ansys

        self._transaction = Ansys.ACT.Mechanical.Transaction()

    def __enter__(self):
        """Enter the context of the transaction."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context of the transaction and disposes of resources."""
        self._transaction.Dispose()



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Miscellaneous utilities."""

import ctypes
import os

TEST_HELPER = None


def _get_test_helper():
    global TEST_HELPER
    import clr

    clr.AddReference("Ans.Common.WB1ManagedUtils")
    import Ansys

    TEST_HELPER = Ansys.Common.WB1ManagedUtils.TestHelper()
    return TEST_HELPER


def sleep(ms: int) -> None:
    """Non-blocking sleep for `ms` milliseconds.

    Mechanical should still work during the sleep.
    """
    _get_test_helper().Wait(ms)


def drain() -> None:
    """Execute all pending work on the main thread.

    Blocks until all the UI messages and other scheduled work complete.
    """
    _get_test_helper().Drain()


def load_library_windows(library: str) -> int:  # pragma: no cover
    """Load a library into the python process on windows."""
    if os.name != "nt":
        return 0

    try:
        load_with_altered_search_path = 8
        dll = ctypes.CDLL(
            library, use_errno=True, use_last_error=True, winmode=load_with_altered_search_path
        )
        return dll._handle
    except OSError:
        return 0



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Shell interaction for PyMechanical embedding."""

import os
import sys
import warnings

import ansys.mechanical.core.embedding.ipython_shell as ipython_shell
import ansys.mechanical.core.embedding.utils as embedding_utils

try:
    import pythoncom

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False


def _install_drain_tracer() -> None:
    warnings.warn("""Mechanical UI without blocking python.
                  The UI will be stuck while python is executing, except during
                  non-blocking sleeps. To use a non-blocking sleep, run
                  `import ansys.mechanical.core.embedding.utils`
                  `ansys.mechanical.core.embedding.utils.sleep(ms)`
                  or, use app.wait_with_dialog() to open a dialog box that
                  will block the python interpreter but keep the UI
                  responsive.""")

    def tracer(frame, event, arg):
        if event == "line":
            from ansys.mechanical.core.embedding.utils import drain

            drain()
        return tracer

    sys.settrace(tracer)


def _uninstall_drain_tracer() -> None:
    sys.settrace(None)


def _use_drain_tracer():
    return os.environ.get("ANSYS_MECHANICAL_EMBEDDING_UI_DRAIN_TRACER") == "1"


def _using_interactive_ipython(warn: bool):
    if not ipython_shell.in_ipython():
        return False

    if not HAS_WIN32:
        if warn:
            warnings.warn("""Interactive PyMechanical requires pywin32""")
        return False

    if os.environ.get("PYMECHANICAL_NO_INTERCTIVE_IPYTHON") == "1":
        return False
    return True


def start_interactive_shell(app):
    """Start the interactive IPython shell."""
    if _using_interactive_ipython(True):

        def _idle_hook():
            embedding_utils.sleep(50)

        def _end_hook():
            app._dispose()

        ipython_shell.get_shell_hooks().idle_hook = _idle_hook
        ipython_shell.get_shell_hooks().end_hook = _end_hook
    else:
        if _use_drain_tracer():
            _install_drain_tracer()
        else:
            warnings.warn("""Interactive PyMechanical is used without IPython.
                          The main thread will be held by python, and therefore
                          the mechanical UI will not be responsive""")


def end_interactive_shell():
    """End the interactive IPython shell."""
    if _using_interactive_ipython(False):
        ipython_shell.get_shell_hooks().idle_hook = None
        ipython_shell.cleanup()
    else:
        if _use_drain_tracer():
            _uninstall_drain_tracer()


def initialize_ipython_shell():
    """Initialize the interactive IPython shell."""
    if not _using_interactive_ipython(False):
        return

    def _start_hook():
        pythoncom.CoInitialize()

    ipython_shell.get_shell_hooks().start_hook = _start_hook
    ipython_shell.post_ipython_blocks()



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Message Manager for App."""

# TODO: add functionality to filter only errors, warnings, info
# TODO: add max number of messages to display
# TODO: implement pep8 formatting

try:  # noqa: F401
    import pandas as pd

    HAS_PANDAS = True
    """Whether or not pandas exists."""
except ImportError:
    HAS_PANDAS = False


class MessageManager:
    """Message manager for adding, fetching, and printing messages."""

    def __init__(self, app):
        """Initialize the message manager."""
        self._app = app

        # Import necessary classes
        from Ansys.Mechanical.Application import Message
        from Ansys.Mechanical.DataModel.Enums import MessageSeverityType

        self._message_severity = MessageSeverityType
        self._message = Message
        self._messages = self._app.ExtAPI.Application.Messages

    def _create_messages_data(self):  # pragma: no cover
        """Update the local cache of messages."""
        data = {
            "Severity": [],
            "TimeStamp": [],
            "DisplayString": [],
            "Source": [],
            "StringID": [],
            "Location": [],
            "RelatedObjects": [],
        }
        for msg in self._app.ExtAPI.Application.Messages:
            data["Severity"].append(str(msg.Severity).upper())
            data["TimeStamp"].append(msg.TimeStamp)
            data["DisplayString"].append(msg.DisplayString)
            data["Source"].append(msg.Source)
            data["StringID"].append(msg.StringID)
            data["Location"].append(msg.Location)
            data["RelatedObjects"].append(msg.RelatedObjects)

        return data

    def __repr__(self):  # pragma: no cover
        """Provide a DataFrame representation of all messages."""
        if not HAS_PANDAS:
            return "Pandas is not available. Please pip install pandas to display messages."
        data = self._create_messages_data()
        return repr(pd.DataFrame(data))

    def __str__(self):
        """Provide a custom string representation of the messages."""
        if self._messages.Count == 0:
            return "No messages to display."

        formatted_messages = [f"[{msg.Severity}] : {msg.DisplayString}" for msg in self._messages]
        return "\n".join(formatted_messages)

    def __getitem__(self, index):
        """Allow indexed access to messages."""
        if len(self._messages) == 0:
            raise IndexError("No messages are available.")
        if index >= len(self._messages) or index < 0:
            raise IndexError("Message index out of range.")
        return self._messages[index]

    def __len__(self):
        """Return the number of messages."""
        return self._messages.Count

    def add(self, severity: str, text: str):
        """Add a message and update the cache.

        Parameters
        ----------
        severity : str
            Severity of the message. Can be "info", "warning", or "error".
        text : str
            Message text.

        Examples
        --------
        >>> app.messages.add("info", "User clicked the start button.")
        """
        severity_map = {
            "info": self._message_severity.Info,
            "warning": self._message_severity.Warning,
            "error": self._message_severity.Error,
        }

        if severity.lower() not in severity_map:
            raise ValueError(f"Invalid severity: {severity}")

        _msg = self._message(text, severity_map[severity.lower()])
        self._messages.Add(_msg)

    def remove(self, index: int):
        """Remove a message by index.

        Parameters
        ----------
        index : int
            Index of the message to remove.

        Examples
        --------
        >>> app.messages.remove(0)
        """
        if index >= len(self._app.ExtAPI.Application.Messages) or index < 0:
            raise IndexError("Message index out of range.")
        _msg = self._messages[index]
        self._messages.Remove(_msg)

    def _show_string(self, filter: str = "Severity;DisplayString") -> str:
        if self._messages.Count == 0:
            return "No messages to display."

        if filter == "*":
            selected_columns = [
                "TimeStamp",
                "Severity",
                "DisplayString",
                "Source",
                "StringID",
                "Location",
                "RelatedObjects",
            ]
        else:
            selected_columns = [col.strip() for col in filter.split(";")]

        lines = []
        for msg in self._messages:
            for key in selected_columns:
                line = f"{key}: {getattr(msg, key, 'Specified attribute not found.')}"
                lines.append(line)
        return "\n".join(lines)

    def show(self, filter="Severity;DisplayString") -> None:
        """Print all messages with full details.

        Parameters
        ----------
        filter : str, optional
            Semicolon separated list of message attributes to display.
            Default is "severity;message".
            if filter is "*", all available attributes will be displayed.

        Examples
        --------
        >>> app.messages.show()
        ... severity: info
        ... message: Sample message.

        >>> app.messages.show(filter="time_stamp;severity;message")
        ... time_stamp: 1/30/2025 12:10:35 PM
        ... severity: info
        ... message: Sample message.
        """
        show_string = self._show_string(filter)
        print(show_string)

    def clear(self):
        """Clear all messages."""
        self._messages.Clear()
